"""
Purpose is to display an image with zooming capabilities.
Reports mouse_events using the event handler f
"""
from enum import Enum
from typing import Tuple, Callable, Dict, List

import numpy as np
from skimage.data import astronaut
from skimage.morphology import skeletonize

from trainer.bib.tgui import pos_from_event, arr_to_pixmap
from PyQt5 import QtWidgets, QtCore, QtGui

from trainer.bib import MaskType


class Brushes(Enum):
    Standard = 1
    AI_Merge = 2


def get_rgb_repr(image_data: np.ndarray,
                 mask: np.ndarray = None,
                 indicator_pos: Tuple[int, int] = None,
                 brush_size=15,
                 frame_number=0) -> np.ndarray:
    if image_data.shape[3] == 1:
        # Assumption: Grayscale
        image_data = image_data[frame_number, :, :, 0]
    else:
        # Assumption: RGB
        image_data = np.dot(image_data[frame_number, :, :, :], [0.2989, 0.5870, 0.1140])

    # if indicator_pos is not None:
    #     indicated = cv2.circle(np.copy(image_data), indicator_pos, brush_size, (255, 0, 0), 1)
    # else:
    #     indicated = image_data
    blend_im = 0.8 * image_data.astype(np.uint8)
    res = np.zeros((image_data.shape[0], image_data.shape[1], 3), dtype=np.uint8)
    if mask is None:
        res[:, :, 0] = blend_im
        res[:, :, 1] = blend_im
        res[:, :, 2] = blend_im
    else:
        res[:, :, 0] = blend_im + 0.2 * (mask.astype(np.uint8) * 255)
        res[:, :, 1] = blend_im
        res[:, :, 2] = blend_im
    return res


class IndicatorSceneItem(QtWidgets.QGraphicsEllipseItem):
    def __init__(self, parent=None, pos: Tuple[int, int] = None, size=15):
        QtWidgets.QGraphicsPixmapItem.__init__(self, parent)

        # self.setFlag(QtWidgets.QGraphicsItem.ItemIsMovable, True)
        # self.setFlag(QtWidgets.QGraphicsItem.ItemIsSelectable, True)
        #
        # # set move restriction rect for the item
        # self.move_restrict_rect = QtCore.QRectF(20, 20, 200, 200)
        # # set item's rectangle
        self.pos, self.size = pos, size
        self.pen = QtGui.QPen(QtCore.Qt.red, 1, QtCore.Qt.SolidLine)
        self.setPen(self.pen)

    def set_position(self, pos: Tuple[int, int]):
        self.pos = pos

    def set_size(self, size: int):
        self.size = size

    def draw(self):
        if self.pos[0] >= 0 and self.pos[1] >= 0:
            self.setRect(self.pos[0] - self.size, self.pos[1] - self.size, self.size * 2, self.size * 2)


class TSegToolGraphicsScene(QtWidgets.QGraphicsScene):
    def __init__(self, f: Callable[[QtGui.QMouseEvent], None], parent=None):
        QtWidgets.QGraphicsScene.__init__(self, parent)

        self.mouse_event_handler = f
        self.pixmap = QtWidgets.QGraphicsPixmapItem()
        self.addItem(self.pixmap)
        self.mask_pixmap = QtWidgets.QGraphicsPixmapItem()
        self.mask_pixmap.setOpacity(0.25)
        self.addItem(self.mask_pixmap)

    def mousePressEvent(self, event: QtGui.QMouseEvent):
        self.mouse_event_handler(event)

    def mouseReleaseEvent(self, event: QtGui.QMouseEvent):
        self.mouse_event_handler(event)

    def mouseMoveEvent(self, event: QtGui.QMouseEvent):
        self.mouse_event_handler(event)


class TSegToolGraphicsView(QtWidgets.QGraphicsView):

    def __init__(self,
                 pen_size_changed: Callable[[int], None],
                 parent=None):
        self.pen_size_changed = pen_size_changed
        QtWidgets.QGraphicsView.__init__(self, parent)
        self.setMouseTracking(True)
        self.setTransformationAnchor(QtWidgets.QGraphicsView.AnchorUnderMouse)

    def set_scene(self, scene: TSegToolGraphicsScene):
        self.setScene(scene)

    def mousePressEvent(self, event):

        if event.button() == QtCore.Qt.MidButton:
            self.setDragMode(QtWidgets.QGraphicsView.ScrollHandDrag)
            self.viewport().setCursor(QtCore.Qt.ClosedHandCursor)
            # self.original_event = event
            handmade_event = QtGui.QMouseEvent(QtCore.QEvent.MouseButtonPress, QtCore.QPointF(event.pos()),
                                               QtCore.Qt.LeftButton, event.buttons(),
                                               QtCore.Qt.KeyboardModifiers())
            self.mousePressEvent(handmade_event)
            # Im Sure I have to do something here.

        super(TSegToolGraphicsView, self).mousePressEvent(event)

    def mouseReleaseEvent(self, event):
        if event.button() == QtCore.Qt.MidButton:
            # for changing back to Qt.OpenHandCursor
            self.viewport().setCursor(QtCore.Qt.OpenHandCursor)
            handmade_event = QtGui.QMouseEvent(QtCore.QEvent.MouseButtonRelease, QtCore.QPointF(event.pos()),
                                               QtCore.Qt.LeftButton,
                                               event.buttons(), QtCore.Qt.KeyboardModifiers())
            self.mouseReleaseEvent(handmade_event)
            self.setDragMode(QtWidgets.QGraphicsView.NoDrag)
        super(TSegToolGraphicsView, self).mouseReleaseEvent(event)

    def center_image(self, scale=True):
        rect = QtCore.QRectF(self.scene().pixmap.pixmap().rect())
        if not rect.isNull():
            self.setSceneRect(rect)
            unity = self.transform().mapRect(QtCore.QRectF(0, 0, 1, 1))
            self.scale(1 / unity.width(), 1 / unity.height())
            viewrect = self.viewport().rect()
            scenerect = self.transform().mapRect(rect)
            factor = min(viewrect.width() / scenerect.width(),
                         viewrect.height() / scenerect.height())
            self.scale(factor, factor)

    def wheelEvent(self, event):
        if event.modifiers() == QtCore.Qt.ControlModifier:
            zoom_in_factor = 1.25
            zoom_out_factor = 1 / zoom_in_factor
            # Zoom
            if event.angleDelta().y() > 0:
                zoom_factor = zoom_in_factor
            else:
                zoom_factor = zoom_out_factor
            self.scale(zoom_factor, zoom_factor)
        elif event.modifiers() == QtCore.Qt.ShiftModifier:
            # Change Brush Size
            self.pen_size_changed(event.angleDelta().y() // 120)
        else:
            return QtWidgets.QGraphicsView.wheelEvent(self, event)


class IndicatorType(Enum):
    Circle = 'circle'


class SegToolController:

    def __init__(self,
                 selection_event_handler: Callable[[List[Tuple[int, int]], bool], None]):
        self._scene = TSegToolGraphicsScene(self.mouse_event_handler)
        self._graphics_view = TSegToolGraphicsView(self.change_pen_size)
        self._graphics_view.set_scene(self._scene)
        self.selection_event_handler = selection_event_handler

        self.pen_size, self._img_stack, self._mask, self._mask_meta = 15, None, None, None
        self.set_img_stack(astronaut().reshape([1, astronaut().shape[0], astronaut().shape[1], 3]))
        self.display_img_stack(frame_number=0)
        self.indicator = IndicatorSceneItem(size=self.pen_size)
        self._scene.addItem(self.indicator)

        self.mouse_down, self.mouse_down_left = False, False
        self.selected_points: List[Tuple[int, int]] = []

    def change_pen_size(self, change: int):
        self.pen_size = self.pen_size + change
        self.indicator.set_size(self.pen_size)
        self.display_indicator()

    def mouse_event_handler(self, e: QtGui.QMouseEvent):
        # Display indicator
        pos = pos_from_event(e)
        self.indicator.set_position(pos)
        self.display_indicator()

        # Handle buttons
        if e.buttons() == QtCore.Qt.LeftButton or e.buttons() == QtCore.Qt.RightButton:
            self.mouse_down_left = e.buttons() == QtCore.Qt.LeftButton
            self.mouse_down = True
            self.selected_points.append(pos)
        elif self.mouse_down:
            self.mouse_down = False
            # Mouse UP!
            self.selection_event_handler(self.selected_points,
                                         add=self.mouse_down_left)
            self.selected_points.clear()

    def get_graphics_scene(self):
        return self._graphics_view

    def set_img_stack(self, img_stack: np.ndarray) -> None:
        self._img_stack = img_stack
        # self._graphics_view.center_image()

    def set_mask(self, mask: np.ndarray, struct_meta: Dict):
        self._mask = mask
        if mask is not None:
            self._mask_meta = struct_meta['meta_data']['structures']

    def display_indicator(self):
        self.indicator.draw()

    def display_img_stack(self, frame_number: int) -> None:
        if self._img_stack.shape[3] == 1:
            # Assumption: Grayscale
            image_data = self._img_stack[frame_number, :, :, 0]
        else:
            # Assumption: RGB
            image_data = np.dot(self._img_stack[frame_number, :, :, :], [0.2989, 0.5870, 0.1140])

        blend_im = image_data.astype(np.uint8)
        res = np.zeros((image_data.shape[0], image_data.shape[1], 3), dtype=np.uint8)
        res[:, :, 0] = blend_im
        res[:, :, 1] = blend_im
        res[:, :, 2] = blend_im

        self._scene.pixmap.setPixmap(arr_to_pixmap(res))

    def display_mask(self, structure_name: str):
        res = np.zeros((self._img_stack.shape[1], self._img_stack.shape[2], 3), dtype=np.uint8)
        if self._mask is not None:
            struct_index = list(self._mask_meta.keys()).index(structure_name)
            struct_mask = self._mask[:, :, struct_index].astype(np.uint8)
            res[:, :, 0] = struct_mask * 255
            if self._mask_meta[structure_name] == MaskType.Line.value:
                res[:, :, 1] = skeletonize(struct_mask) * 255
            elif self._mask_meta[structure_name] == MaskType.Point.value:
                raise NotImplementedError()
            res[:, :, 2] = np.zeros((self._mask.shape[0], self._mask.shape[1]), dtype=np.uint8)
            for i in range(len(self._mask_meta.keys())):
                if i != struct_index:
                    res[:, :, 2] |= (self._mask[:, :, i].astype(np.uint8) * 255)
            # res[:, :, 2] = self._mask[:, :, i].astype(np.uint8) * 255
        self._scene.mask_pixmap.setPixmap(arr_to_pixmap(res))



# Copyright 2019 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Pix2pix.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os
import time
from absl import app
from absl import flags
import tensorflow as tf  # TF2

assert tf.__version__.startswith('2')

FLAGS = flags.FLAGS

flags.DEFINE_integer('buffer_size', 400, 'Shuffle buffer size')
flags.DEFINE_integer('batch_size', 1, 'Batch Size')
flags.DEFINE_integer('epochs', 1, 'Number of epochs')
flags.DEFINE_string('path', None, 'Path to the data folder')
flags.DEFINE_boolean('enable_function', True, 'Enable Function?')

IMG_WIDTH = 256
IMG_HEIGHT = 256
AUTOTUNE = tf.data.experimental.AUTOTUNE


def load(image_file):
    """Loads the image and generates input and target image.

    Args:
      image_file: .jpeg file

    Returns:
      Input image, target image
    """
    image = tf.io.read_file(image_file)
    image = tf.image.decode_jpeg(image)

    w = tf.shape(image)[1]

    w = w // 2
    real_image = image[:, :w, :]
    input_image = image[:, w:, :]

    input_image = tf.cast(input_image, tf.float32)
    real_image = tf.cast(real_image, tf.float32)

    return input_image, real_image


def resize(input_image, real_image, height, width):
    input_image = tf.image.resize(input_image, [height, width],
                                  method=tf.image.ResizeMethod.NEAREST_NEIGHBOR)
    real_image = tf.image.resize(real_image, [height, width],
                                 method=tf.image.ResizeMethod.NEAREST_NEIGHBOR)

    return input_image, real_image


def random_crop(input_image, real_image):
    stacked_image = tf.stack([input_image, real_image], axis=0)
    cropped_image = tf.image.random_crop(
        stacked_image, size=[2, IMG_HEIGHT, IMG_WIDTH, 3])

    return cropped_image[0], cropped_image[1]


def normalize(input_image, real_image):
    input_image = (input_image / 127.5) - 1
    real_image = (real_image / 127.5) - 1

    return input_image, real_image


@tf.function
def random_jitter(input_image, real_image):
    """Random jittering.

    Resizes to 286 x 286 and then randomly crops to IMG_HEIGHT x IMG_WIDTH.

    Args:
      input_image: Input Image
      real_image: Real Image

    Returns:
      Input Image, real image
    """
    # resizing to 286 x 286 x 3
    input_image, real_image = resize(input_image, real_image, 286, 286)

    # randomly cropping to 256 x 256 x 3
    input_image, real_image = random_crop(input_image, real_image)

    if tf.random.uniform(()) > 0.5:
        # random mirroring
        input_image = tf.image.flip_left_right(input_image)
        real_image = tf.image.flip_left_right(real_image)

    return input_image, real_image


def load_image_train(image_file):
    input_image, real_image = load(image_file)
    input_image, real_image = random_jitter(input_image, real_image)
    input_image, real_image = normalize(input_image, real_image)

    return input_image, real_image


def load_image_test(image_file):
    input_image, real_image = load(image_file)
    input_image, real_image = resize(input_image, real_image,
                                     IMG_HEIGHT, IMG_WIDTH)
    input_image, real_image = normalize(input_image, real_image)

    return input_image, real_image


def create_dataset(path_to_train_images, path_to_test_images, buffer_size,
                   batch_size):
    """Creates a tf.data Dataset.

    Args:
      path_to_train_images: Path to train images folder.
      path_to_test_images: Path to test images folder.
      buffer_size: Shuffle buffer size.
      batch_size: Batch size

    Returns:
      train dataset, test dataset
    """
    train_dataset = tf.data.Dataset.list_files(path_to_train_images)
    train_dataset = train_dataset.shuffle(buffer_size)
    train_dataset = train_dataset.map(
        load_image_train, num_parallel_calls=AUTOTUNE)
    train_dataset = train_dataset.batch(batch_size)

    test_dataset = tf.data.Dataset.list_files(path_to_test_images)
    test_dataset = test_dataset.map(
        load_image_test, num_parallel_calls=AUTOTUNE)
    test_dataset = test_dataset.batch(batch_size)

    return train_dataset, test_dataset


class InstanceNormalization(tf.keras.layers.Layer):
    """Instance Normalization Layer (https://arxiv.org/abs/1607.08022)."""

    def __init__(self, epsilon=1e-5):
        super(InstanceNormalization, self).__init__()
        self.epsilon = epsilon

    def build(self, input_shape):
        self.scale = self.add_weight(
            name='scale',
            shape=input_shape[-1:],
            initializer=tf.random_normal_initializer(1., 0.02),
            trainable=True)

        self.offset = self.add_weight(
            name='offset',
            shape=input_shape[-1:],
            initializer='zeros',
            trainable=True)

    def call(self, x):
        mean, variance = tf.nn.moments(x, axes=[1, 2], keepdims=True)
        inv = tf.math.rsqrt(variance + self.epsilon)
        normalized = (x - mean) * inv
        return self.scale * normalized + self.offset


def downsample(filters, size, norm_type='batchnorm', apply_norm=True):
    """Downsamples an input.

    Conv2D => Batchnorm => LeakyRelu

    Args:
      filters: number of filters
      size: filter size
      norm_type: Normalization type; either 'batchnorm' or 'instancenorm'.
      apply_norm: If True, adds the batchnorm layer

    Returns:
      Downsample Sequential Model
    """
    initializer = tf.random_normal_initializer(0., 0.02)

    result = tf.keras.Sequential()
    result.add(
        tf.keras.layers.Conv2D(filters, size, strides=2, padding='same',
                               kernel_initializer=initializer, use_bias=False))

    if apply_norm:
        if norm_type.lower() == 'batchnorm':
            result.add(tf.keras.layers.BatchNormalization())
        elif norm_type.lower() == 'instancenorm':
            result.add(InstanceNormalization())

    result.add(tf.keras.layers.LeakyReLU())

    return result


def upsample(filters, size, norm_type='batchnorm', apply_dropout=False):
    """Upsamples an input.

    Conv2DTranspose => Batchnorm => Dropout => Relu

    Args:
      filters: number of filters
      size: filter size
      norm_type: Normalization type; either 'batchnorm' or 'instancenorm'.
      apply_dropout: If True, adds the dropout layer

    Returns:
      Upsample Sequential Model
    """

    initializer = tf.random_normal_initializer(0., 0.02)

    result = tf.keras.Sequential()
    result.add(
        tf.keras.layers.Conv2DTranspose(filters, size, strides=2,
                                        padding='same',
                                        kernel_initializer=initializer,
                                        use_bias=False))

    if norm_type.lower() == 'batchnorm':
        result.add(tf.keras.layers.BatchNormalization())
    elif norm_type.lower() == 'instancenorm':
        result.add(InstanceNormalization())

    if apply_dropout:
        result.add(tf.keras.layers.Dropout(0.5))

    result.add(tf.keras.layers.ReLU())

    return result


def unet_generator(output_channels, norm_type='batchnorm'):
    """Modified u-net generator model (https://arxiv.org/abs/1611.07004).

    Args:
      output_channels: Output channels
      norm_type: Type of normalization. Either 'batchnorm' or 'instancenorm'.

    Returns:
      Generator model
    """

    down_stack = [
        downsample(64, 4, norm_type, apply_norm=False),  # (bs, 128, 128, 64)
        downsample(128, 4, norm_type),  # (bs, 64, 64, 128)
        downsample(256, 4, norm_type),  # (bs, 32, 32, 256)
        downsample(512, 4, norm_type),  # (bs, 16, 16, 512)
        downsample(512, 4, norm_type),  # (bs, 8, 8, 512)
        downsample(512, 4, norm_type),  # (bs, 4, 4, 512)
        downsample(512, 4, norm_type),  # (bs, 2, 2, 512)
        downsample(512, 4, norm_type),  # (bs, 1, 1, 512)
    ]

    up_stack = [
        upsample(512, 4, norm_type, apply_dropout=True),  # (bs, 2, 2, 1024)
        upsample(512, 4, norm_type, apply_dropout=True),  # (bs, 4, 4, 1024)
        upsample(512, 4, norm_type, apply_dropout=True),  # (bs, 8, 8, 1024)
        upsample(512, 4, norm_type),  # (bs, 16, 16, 1024)
        upsample(256, 4, norm_type),  # (bs, 32, 32, 512)
        upsample(128, 4, norm_type),  # (bs, 64, 64, 256)
        upsample(64, 4, norm_type),  # (bs, 128, 128, 128)
    ]

    initializer = tf.random_normal_initializer(0., 0.02)
    last = tf.keras.layers.Conv2DTranspose(
        output_channels, 4, strides=2,
        padding='same', kernel_initializer=initializer,
        activation='tanh')  # (bs, 256, 256, 3)

    concat = tf.keras.layers.Concatenate()

    inputs = tf.keras.layers.Input(shape=[None, None, 3])
    x = inputs

    # Downsampling through the model
    skips = []
    for down in down_stack:
        x = down(x)
        skips.append(x)

    skips = reversed(skips[:-1])

    # Upsampling and establishing the skip connections
    for up, skip in zip(up_stack, skips):
        x = up(x)
        x = concat([x, skip])

    x = last(x)

    return tf.keras.Model(inputs=inputs, outputs=x)


def discriminator(norm_type='batchnorm', target=True):
    """PatchGan discriminator model (https://arxiv.org/abs/1611.07004).

    Args:
      norm_type: Type of normalization. Either 'batchnorm' or 'instancenorm'.
      target: Bool, indicating whether target image is an input or not.

    Returns:
      Discriminator model
    """

    initializer = tf.random_normal_initializer(0., 0.02)

    inp = tf.keras.layers.Input(shape=[None, None, 3], name='input_image')
    x = inp

    if target:
        tar = tf.keras.layers.Input(shape=[None, None, 3], name='target_image')
        x = tf.keras.layers.concatenate([inp, tar])  # (bs, 256, 256, channels*2)

    down1 = downsample(64, 4, norm_type, False)(x)  # (bs, 128, 128, 64)
    down2 = downsample(128, 4, norm_type)(down1)  # (bs, 64, 64, 128)
    down3 = downsample(256, 4, norm_type)(down2)  # (bs, 32, 32, 256)

    zero_pad1 = tf.keras.layers.ZeroPadding2D()(down3)  # (bs, 34, 34, 256)
    conv = tf.keras.layers.Conv2D(
        512, 4, strides=1, kernel_initializer=initializer,
        use_bias=False)(zero_pad1)  # (bs, 31, 31, 512)

    if norm_type.lower() == 'batchnorm':
        norm1 = tf.keras.layers.BatchNormalization()(conv)
    elif norm_type.lower() == 'instancenorm':
        norm1 = InstanceNormalization()(conv)

    leaky_relu = tf.keras.layers.LeakyReLU()(norm1)

    zero_pad2 = tf.keras.layers.ZeroPadding2D()(leaky_relu)  # (bs, 33, 33, 512)

    last = tf.keras.layers.Conv2D(
        1, 4, strides=1,
        kernel_initializer=initializer)(zero_pad2)  # (bs, 30, 30, 1)

    if target:
        return tf.keras.Model(inputs=[inp, tar], outputs=last)
    else:
        return tf.keras.Model(inputs=inp, outputs=last)


def get_checkpoint_prefix():
    checkpoint_dir = './training_checkpoints'
    checkpoint_prefix = os.path.join(checkpoint_dir, 'ckpt')

    return checkpoint_prefix


class Pix2pix(object):
    """Pix2pix class.

    Args:
      epochs: Number of epochs.
      enable_function: If true, train step is decorated with tf.function.
      buffer_size: Shuffle buffer size..
      batch_size: Batch size.
    """

    def __init__(self, epochs, enable_function):
        self.epochs = epochs
        self.enable_function = enable_function
        self.lambda_value = 100
        self.loss_object = tf.keras.losses.BinaryCrossentropy(from_logits=True)
        self.generator_optimizer = tf.keras.optimizers.Adam(2e-4, beta_1=0.5)
        self.discriminator_optimizer = tf.keras.optimizers.Adam(2e-4, beta_1=0.5)
        self.generator = unet_generator(output_channels=3)
        self.discriminator = discriminator()
        self.checkpoint = tf.train.Checkpoint(
            generator_optimizer=self.generator_optimizer,
            discriminator_optimizer=self.discriminator_optimizer,
            generator=self.generator,
            discriminator=self.discriminator)

    def discriminator_loss(self, disc_real_output, disc_generated_output):
        real_loss = self.loss_object(
            tf.ones_like(disc_real_output), disc_real_output)

        generated_loss = self.loss_object(tf.zeros_like(
            disc_generated_output), disc_generated_output)

        total_disc_loss = real_loss + generated_loss

        return total_disc_loss

    def generator_loss(self, disc_generated_output, gen_output, target):
        gan_loss = self.loss_object(tf.ones_like(
            disc_generated_output), disc_generated_output)

        # mean absolute error
        l1_loss = tf.reduce_mean(tf.abs(target - gen_output))
        total_gen_loss = gan_loss + (self.lambda_value * l1_loss)
        return total_gen_loss

    def train_step(self, input_image, target_image):
        """One train step over the generator and discriminator model.

        Args:
          input_image: Input Image.
          target_image: Target image.

        Returns:
          generator loss, discriminator loss.
        """
        with tf.GradientTape() as gen_tape, tf.GradientTape() as disc_tape:
            gen_output = self.generator(input_image, training=True)

            disc_real_output = self.discriminator(
                [input_image, target_image], training=True)
            disc_generated_output = self.discriminator(
                [input_image, gen_output], training=True)

            gen_loss = self.generator_loss(
                disc_generated_output, gen_output, target_image)
            disc_loss = self.discriminator_loss(
                disc_real_output, disc_generated_output)

        generator_gradients = gen_tape.gradient(
            gen_loss, self.generator.trainable_variables)
        discriminator_gradients = disc_tape.gradient(
            disc_loss, self.discriminator.trainable_variables)

        self.generator_optimizer.apply_gradients(zip(
            generator_gradients, self.generator.trainable_variables))
        self.discriminator_optimizer.apply_gradients(zip(
            discriminator_gradients, self.discriminator.trainable_variables))

        return gen_loss, disc_loss

    def train(self, dataset, checkpoint_pr):
        """Train the GAN for x number of epochs.

        Args:
          dataset: train dataset.
          checkpoint_pr: prefix in which the checkpoints are stored.

        Returns:
          Time for each epoch.
        """
        time_list = []
        if self.enable_function:
            self.train_step = tf.function(self.train_step)

        for epoch in range(self.epochs):
            start_time = time.time()
            for input_image, target_image in dataset:
                gen_loss, disc_loss = self.train_step(input_image, target_image)

            wall_time_sec = time.time() - start_time
            time_list.append(wall_time_sec)

            # saving (checkpoint) the model every 20 epochs
            if (epoch + 1) % 20 == 0:
                self.checkpoint.save(file_prefix=checkpoint_pr)

            template = 'Epoch {}, Generator loss {}, Discriminator Loss {}'
            print(template.format(epoch, gen_loss, disc_loss))

        return time_list


def run_main(argv):
    del argv
    kwargs = {'epochs': FLAGS.epochs, 'enable_function': FLAGS.enable_function,
              'path': FLAGS.path, 'buffer_size': FLAGS.buffer_size,
              'batch_size': FLAGS.batch_size}
    main(**kwargs)


def main(epochs, enable_function, path, buffer_size, batch_size):
    path_to_folder = path

    pix2pix_object = Pix2pix(epochs, enable_function)

    train_dataset, _ = create_dataset(
        os.path.join(path_to_folder, 'train/*.jpg'),
        os.path.join(path_to_folder, 'test/*.jpg'),
        buffer_size, batch_size)
    checkpoint_pr = get_checkpoint_prefix()
    print('Training ...')
    return pix2pix_object.train(train_dataset, checkpoint_pr)


if __name__ == '__main__':
    app.run(run_main)



import pydicom
from typing import List

ALLOWED_TYPES = [
    str,
    List[str],
    int,
    float
]


def import_dicom(dicom_path: str):
    ds = pydicom.dcmread(dicom_path)
    ds.decompress()

    if 'PixelData' in ds:

        img_data = ds.pixel_array

        meta = {}
        for key in ds.keys():
            v = ds[key]
            try:  # Use try to check if v is even subscriptable
                key, val = v.keyword, v.value

                if type(val) in ALLOWED_TYPES:
                    if key and val:  # Checks for empty strings
                        meta[key] = val
                        print(f"Added {key} with value {val}")
                    else:
                        print(f"{key} seems to be empty, skip!")
                else:
                    print(f"{key} has the wrong type: {type(val)}")
            except Exception as e:
                print(e)
                print(f"ignored: {str(v)}")
    else:
        raise Exception("The dicom file seems to not contain any pixel data")

    return img_data, meta



from trainer.ml import Dataset
# from ml.models.unet_mod import compile_and_train

import tensorflow as tf
import numpy as np
from matplotlib import pyplot as plt
import seaborn as sns
from trainer.ml.models.deeplab import get_model, predict as dl_predict, compile_and_train as dl_compile_and_train

print("INITIALIZED ALL AI STUFF")

model = get_model(tf.keras.utils.get_file('bone_segmentation_binary_weights.h5',
                                          'https://rwth-aachen.sciebo.de/s/s7wR70RG85PMwM6/download'))


def compile_and_train(ds: Dataset, structure_name='bone'):
    return dl_compile_and_train(ds, model, structure_name=structure_name)


def predict(im: np.ndarray, ds: Dataset, structure_name: str):
    return dl_predict(model, im, ds, structure_name)


if __name__ == '__main__':
    import imageio

    im = imageio.imread("../sample_data/example_image.png")
    result = predict(im)
    sns.heatmap(result)
    plt.show()
    plt.imshow(im, cmap=plt.cm.gray)
    plt.show()



import numpy as np
from PyQt5 import QtGui, QtWidgets


def pos_from_event(e):
    return int(e.scenePos().x()), int(e.scenePos().y())


def arr_to_pixmap(arr: np.ndarray) -> QtGui.QPixmap:
    image = QtGui.QImage(
        arr, arr.shape[1], arr.shape[0], arr.shape[1] * 3, QtGui.QImage.Format_RGB888)
    return QtGui.QPixmap(image)


def open_file_dialog(headline="Select a file", parent=None, dir_name=None):
    """Select a file via a dialog and return the file name."""
    if dir_name is None: dir_name = '/'
    fname = QtWidgets.QFileDialog.getOpenFileName(parent, headline,
                                                  dir_name, filter="All files (*);; SM Files (*.sm)")
    return fname[0]


def open_folder_dialog(parent=None, headline="Select directory") -> str:
    return str(QtWidgets.QFileDialog.getExistingDirectory(parent, headline))



from enum import Enum
from typing import Generator, Tuple, Iterable, Dict

import numpy as np
import matplotlib.pyplot as plt
import seaborn as sns
import cv2
from scipy.ndimage.morphology import distance_transform_edt as dist_trans

from trainer.ml import Subject, Dataset
from trainer.bib import create_identifier


class ImageNormalizations(Enum):
    UnitRange = 1


def append_dicom_to_subject(te_path: str,
                            dicom_path: str,
                            binary_name: str = '',
                            seg_structs: Dict[str, str] = None,
                            auto_save=True) -> Subject:
    """

    :param te_path: directory path to the subject
    :param dicom_path: filepath to the dicom containing the image data
    :param binary_name: Name of the binary, if not provided a name is chosen.
    :param seg_structs: Structures that can be segmented in the image data
    :param auto_save: The new state of the subject is automatically saved to disk
    :return: The subject containing the new data
    """
    s = Subject.from_disk(te_path)

    if not binary_name:
        binary_name = create_identifier(hint='DICOM')

    from trainer.bib.dicom_utils import import_dicom

    img_data, meta = import_dicom(dicom_path)
    s.add_source_image_by_arr(img_data, binary_name, structures=seg_structs, extra_info=meta)

    if auto_save:
        s.to_disk(s.get_parent_directory())

    return s


def normalize_im(im: np.ndarray, norm_type=ImageNormalizations.UnitRange):
    """
    Currently just normalizes an image with pixel intensities in range [0, 255] to [-1, 1]
    :return: The normalized image
    """
    if norm_type == ImageNormalizations.UnitRange:
        return (im.astype(np.float32) / 127.5) - 1
    else:
        raise Exception("Unknown Normalization type")


def gray_to_rgb(im):
    im_stacked = np.zeros((im.shape[0], im.shape[1], 3))
    im_stacked[:, :, 0] = im
    im_stacked[:, :, 1] = im
    im_stacked[:, :, 2] = im
    return im_stacked


def append_dim(g: Iterable[Tuple[np.ndarray, np.ndarray]]) -> Iterable[Tuple[np.ndarray, np.ndarray]]:
    f = lambda arr: np.expand_dims(arr, axis=len(arr.shape))
    for im, gt in g:
        yield f(im), f(gt)


def distance_transformed(g: Generator) -> Generator:
    for im, gt in g:
        yield im, dist_trans(np.invert(gt.astype(np.bool)).astype(np.float32))


def resize(g: Iterable[Tuple[np.ndarray, np.ndarray]], s: Tuple[int, int]) -> Iterable[Tuple[np.ndarray, np.ndarray]]:
    for im, gt in g:
        im_resized = cv2.resize(im, s)
        gt_resized = cv2.resize(gt, s)
        yield im_resized, gt_resized


def batcherize(g: Iterable[Tuple[np.ndarray, np.ndarray]], batchsize=8) -> Iterable[Tuple[np.ndarray, np.ndarray]]:
    """
    Note that all images have to be of the same size.
    :param g:
    :param batchsize:
    :return:
    """
    ims, gts = [], []
    for im, gt in g:
        if len(ims) == batchsize:
            res = np.array(ims), np.array(gts)
            yield res
            ims, gts = [], []
        ims.append(im)
        gts.append(gt)


def extract_segmentation_pair(g: Iterable[Subject], src_name: str) -> Iterable[Tuple[np.ndarray, np.ndarray]]:
    for te in g:
        res = te.get_grayimage_training_tuple_raw(src_name)
        yield res


def pair_augmentation(g: Iterable[Tuple[np.ndarray, np.ndarray]], aug_ls) -> Iterable[Tuple[np.ndarray, np.ndarray]]:
    import imgaug.augmenters as iaa
    seq = iaa.Sequential(aug_ls)
    for im, gt in g:
        images_aug = seq(images=[im], segmentation_maps=[gt.astype(np.bool)])
        res = images_aug[0][0].astype(np.float32), images_aug[1][0][:, :, 0].astype(np.float32)
        yield res


def visualize_batch(t: Tuple[np.ndarray, np.ndarray]) -> None:
    im_arr, gt_arr = t
    ims = [im_arr[i, :, :, 0] for i in range(im_arr.shape[0])]
    gts = [gt_arr[i, :, :, 0] for i in range(gt_arr.shape[0])]
    for i, _ in enumerate(ims):
        plt.subplot(1, 2, 1)
        sns.heatmap(ims[i])
        plt.subplot(1, 2, 2)
        sns.heatmap(gts[i])
        plt.show()


if __name__ == '__main__':
    d = Dataset.from_disk("C:\\Users\\rapha\\Desktop\\dataset_folder\\all_manual_train")

    g = d.random_subject_generator(split="train")
    g = extract_segmentation_pair(g)

    g = pair_augmentation(g, [
        iaa.Crop(px=(1, 16), keep_size=False),
        iaa.Fliplr(0.5),
        iaa.GammaContrast((0.5, 1.5)),
        iaa.Sometimes(0.5,
                      iaa.GaussianBlur(sigma=(0, 3.0))
                      ),
    ])
    g = resize(g, (384, 384))
    g = append_dim(batcherize(g))
    o = next(g)



import tensorflow as tf


def focal_loss(alpha=0.25, gamma=2):
    def focal_loss_with_logits(logits, targets, alpha, gamma, y_pred):
        weight_a = alpha * (1 - y_pred) ** gamma * targets
        weight_b = (1 - alpha) * y_pred ** gamma * (1 - targets)

        return (tf.math.log1p(tf.exp(-tf.abs(logits))) + tf.nn.relu(-logits)) * (
                weight_a + weight_b) + logits * weight_b

    def loss(y_true, y_pred):
        y_pred = tf.clip_by_value(y_pred, tf.keras.backend.epsilon(), 1 - tf.keras.backend.epsilon())
        logits = tf.math.log(y_pred / (1 - y_pred))

        loss = focal_loss_with_logits(logits=logits, targets=y_true, alpha=alpha, gamma=gamma, y_pred=y_pred)

        return tf.reduce_mean(loss)

    return loss



import sys
import os
from typing import Tuple, Dict, Callable

from PyQt5 import QtCore, QtWidgets, QtGui
from PyQt5.QtCore import Qt


class TContentGrid(QtWidgets.QWidget):
    def __init__(self, main_widget: QtWidgets.QWidget):
        super(TContentGrid, self).__init__()
        layout = QtWidgets.QHBoxLayout()

        # Toolbar
        toolbar_scroller = QtWidgets.QScrollArea()
        toolbar_scroller.setFixedWidth(250)
        toolbar_scroller.setWidgetResizable(True)
        toolbar_widget = QtWidgets.QWidget()
        self.toolbar = QtWidgets.QVBoxLayout()
        self.toolbar.setAlignment(Qt.AlignTop)
        toolbar_widget.setLayout(self.toolbar)
        toolbar_scroller.setWidget(toolbar_widget)
        layout.addWidget(toolbar_scroller)

        # Main Content
        right_side = QtWidgets.QWidget()
        right_side_layout = QtWidgets.QVBoxLayout()
        # self.main_content = QtWidgets.QWidget()
        self.main_content = main_widget

        # Bottom
        self.logging_widget = QtWidgets.QPlainTextEdit("Here we will see a process log in the future")
        self.logging_widget.setReadOnly(True)
        self.logging_widget.setFrameStyle(QtWidgets.QFrame.StyledPanel)

        logging_splitter = QtWidgets.QSplitter(Qt.Vertical)
        right_side_layout.addWidget(logging_splitter)
        logging_splitter.addWidget(self.main_content)
        # logging_splitter.addWidget(self.logging_widget)
        logging_splitter.setSizes([600, 200])

        right_side.setLayout(right_side_layout)
        layout.addWidget(right_side)

        self.setLayout(layout)

    def add_tool(self, w):
        self.toolbar.addWidget(w)


class TWindow(QtWidgets.QMainWindow):

    def __init__(self, main_widget: QtWidgets.QWidget = None, title="Window title", maximized=False):
        super(TWindow, self).__init__()
        if main_widget is not None:
            self.main_widget = main_widget
        else:
            self.main_widget = QtWidgets.QLabel("No main content is set")
        self.actions = {"Window": [
            self.create_action_from_tuple("Close Window", "Ctrl+Q", "Closes the current window", self.exit_window)
        ]}
        self.sub_windows: Dict[str, TWindow] = {}

        self.setWindowTitle(title)
        self.content_grid = None
        self.gui_initialized = False
        if maximized:
            self.showMaximized()

    def init_ui(self, actions: Dict = None):
        if self.gui_initialized:
            return
        if actions is not None:
            for action_key in actions:
                self.actions[action_key] = actions[action_key]

        self.statusBar()
        self.create_menu_bar()
        self.content_grid = TContentGrid(self.main_widget)
        self.setCentralWidget(self.content_grid)
        self.gui_initialized = True

    def show_subwindow(self, name: str, w):
        self.sub_windows[name] = w
        w.show()

    def create_action_from_tuple(self, action_name: str, shortcut: str, status_tip: str, handler: Callable):
        action = QtWidgets.QAction(f"&{action_name}", self)
        action.setShortcut(shortcut)
        action.setStatusTip(status_tip)
        action.triggered.connect(handler)
        return action

    def exit_window(self):
        self.close()

    def create_menu_bar(self):
        main_menu = self.menuBar()
        for parent_key in self.actions:
            menu = main_menu.addMenu(f"&{parent_key}")
            for a in self.actions[parent_key]:
                menu.addAction(a)

    def get_user_input(self, headline="Headline", dialog_text="Input"):
        return QtWidgets.QInputDialog.getText(self, headline, dialog_text)


def run_window(cls, *args):
    def exception_hook(exctype, value, traceback):
        print(exctype, value, traceback)
        sys._excepthook(exctype, value, traceback)
        sys.exit(1)

    sys._excepthook = sys.excepthook
    sys.excepthook = exception_hook

    app = QtWidgets.QApplication(sys.argv)
    gui = cls(*args)
    gui.init_ui()
    gui.show()
    sys.exit(app.exec_())


if __name__ == '__main__':
    run_window(TWindow)



import re
import urllib
import tempfile
import zipfile
import os
from typing import Tuple, List, Dict
from pathlib import Path

import numpy as np
import cv2
import imageio
import PySimpleGUI as sg
from datetime import datetime


def create_identifier(hint: str = '') -> str:
    """
    Can be used to create unique names for files.
    Follows the form YYYY_MM_DD__hh_mm_ss
    :return:
    """
    now = datetime.now()
    dt_string = now.strftime("%Y_%m_%d__%H_%M_%S")
    return f"{dt_string}_{hint}" if hint else dt_string


def standalone_foldergrab(
        folder_not_file: bool = False,
        optional_inputs: List[Tuple[str, str]] = None,
        optional_choices: List[Tuple[str, str, List[str]]] = None,
        title="Select a folder") -> Tuple[str, Dict[str, str]]:
    """
    Uses a simple blocking GUI for prompting the user for a file or folder path.
    Optionally allows to prompt for additional text inputs as well.
    :param folder_not_file: if True asks for a folder, if False asks for a file
    :param optional_inputs: A list of (description, key) pairs
    :param optional_choices: A list of (description, key, list of options) pairs
    :param title:
    :return: Tuple with path as first entry and a dictionary with the optional text inputs as second entry
    """
    if optional_inputs is None:
        optional_inputs = []
    if optional_choices is None:
        optional_choices = []
    layout = [[sg.Text(text=title)],
              [sg.Input(key='folder_name'), sg.FolderBrowse() if folder_not_file else sg.FileBrowse()]]
    for t in optional_inputs:
        new_item = [sg.Text(text=t[0], size=(15, 1)), sg.InputText(key=t[1])]
        layout.append(new_item)
    for t in optional_choices:
        new_item = [sg.Text(text=t[0], size=(15, 1)), sg.Combo(t[2], key=t[1])]
        layout.append(new_item)
    layout.append([sg.Submit(), sg.Cancel()])

    p = sg.Window(title, layout)
    e, v = p.Read()
    if e is None or e == 'Cancel':
        p.close()
        return "", {}
    elif e == 'Submit':
        p.close()
        res = {t[1]: v[t[1]] for t in optional_inputs}
        res.update({t[1]: v[t[1]] for t in optional_choices})
        return v['folder_name'], res


def make_converter_dict_for_enum(e):
    res = {v: v.value for v in e}
    res.update({v.value: v for v in e})
    return res


def load_grayscale_from_disk(path: str) -> np.ndarray:
    im = imageio.imread(path)
    if len(im.shape) > 2:
        im = cv2.cvtColor(im, cv2.COLOR_RGB2GRAY)
    return im


def download_and_extract(online_url: str, parent_dir='./', dir_name: str = None) -> str:
    """
    Can be used to download and extract a .zip dataset file hosted online.
    Assumes the zip to be one directory.
    :param online_url: The url that points directly to a .zip file containing folders with the dataset files
    :param parent_dir: The directory that is used to store the temporary zip and the final extracted folder
    :param dir_name: if provided, the function checks if the directory already exists and does not download it again
    :return: The absolute local path to the directory
    """
    # Check for existence
    if dir_name is not None:
        abs_dirname = os.path.join(parent_dir, dir_name)
        if os.path.exists(abs_dirname):
            print(f"The local copy already exists: {online_url} at {abs_dirname}")
            return abs_dirname
    else:
        print(f"The local copy does not yet exist, proceed to download it")

    # Download
    print(f"Starting downloading from {online_url}")
    temp_file = tempfile.NamedTemporaryFile(mode='w')
    temp_file.close()
    urllib.request.urlretrieve(online_url, temp_file.name)

    # Extract
    print(f"Start extracting {temp_file.name}")
    zf = zipfile.ZipFile(temp_file.name)
    zf.extractall(path=parent_dir)
    zf.close()

    first_filename = zf.filelist[0].filename
    res = os.path.join(parent_dir, str(Path(first_filename).parent))
    if dir_name is not None and dir_name not in res:
        raise Exception(f"The provided directory name {dir_name} did not match the actual directory name {res}")
    return res


def slugify(value):
    """
    Normalizes string, converts to lowercase, removes non-alpha characters,
    and converts spaces to hyphens.
    """
    return re.sub(r'[\\/*?:"<>|]', "", value)


if __name__ == '__main__':
    folder, x = standalone_foldergrab(folder_not_file=True, optional_inputs=[("Download Path: ", "d_p")],
                                      title="Download")
    download_url = x['d_p']
    p = download_and_extract(download_url, parent_dir=folder)
    print("finished")



"""
A server that is concerned with the manipulation of one dataset at a time.
"""

from flask import Flask, jsonify, render_template

from trainer import VERSION

app = Flask(__name__)
print(__name__)


@app.route('/')
def angular_frontend():
    return render_template(template_name_or_list='main.html', version=VERSION)


@app.route('/version')
def get_version():
    return {
        "version": VERSION
    }



from typing import Callable
from enum import Enum

from PyQt5 import QtWidgets, QtCore
import PySimpleGUI as sg

from trainer.ml import Dataset, Subject
from trainer.bib import ClassSelectionLevel


class TClassBox(QtWidgets.QWidget):
    """
    Takes the information of one class.
    """

    def __init__(self, class_info, callback: Callable[[str, str], None], class_name=None):
        super().__init__()
        self.f_changed = callback
        self.class_name = class_name

        self.label = QtWidgets.QLabel(f"Set Class: {class_name}")

        self.selector = QtWidgets.QComboBox()
        self.selector.addItem("--Removed--")
        self.selector.addItems(class_info['values'])
        self.selector.activated[str].connect(self.selection_changed)

        self.layout = QtWidgets.QVBoxLayout()
        self.layout.addWidget(self.label)
        self.layout.addWidget(self.selector)
        self.setLayout(self.layout)

    def selection_changed(self, text):
        self.f_changed(self.class_name, text)

    def update_values(self, s: Subject, for_binary=""):
        index = self.selector.findText(s.get_class_value(self.class_name, for_binary=for_binary),
                                       QtCore.Qt.MatchFixedString)
        if index >= 0:
            self.selector.setCurrentIndex(index)


class TClassSelector(QtWidgets.QWidget):
    """
    Container for multiple classes, instantiates a TClassBox for every class
    """

    def __init__(self):
        super().__init__()
        self.layout = QtWidgets.QVBoxLayout()
        self.class_boxes = []
        self.subject, self.binary_name, self.frame_number = None, '', -1
        self.selection_level = ClassSelectionLevel.SubjectLevel

        self.label = QtWidgets.QLabel("No Selected Subject")

        self.level_selector = QtWidgets.QComboBox()
        self.level_selector.addItems([selection_level.value for selection_level in ClassSelectionLevel])
        self.level_selector.activated[str].connect(self.selection_level_changed)

        self.layout.addWidget(self.label)
        self.layout.addWidget(self.level_selector)
        self.setLayout(self.layout)

    def selection_level_changed(self, text):
        next_selection_level = ClassSelectionLevel.SubjectLevel
        for selection_level in ClassSelectionLevel:
            if text == selection_level.value:
                next_selection_level = selection_level

        if next_selection_level == ClassSelectionLevel.SubjectLevel:
            pass
        elif next_selection_level == ClassSelectionLevel.BinaryLevel:
            pass
        elif next_selection_level == ClassSelectionLevel.FrameLevel:
            sg.popup("Frame level class selection not supported yet, go back and hope for the best")

        self.selection_level = next_selection_level
        self.update_label()

    def configure_selection(self, d: Dataset):
        class_names = d.get_class_names()

        def change_handler(class_name: str, class_value: str):
            print(class_name)
            print(class_value)
            if class_value == '--Removed--':
                if self.selection_level == ClassSelectionLevel.BinaryLevel:
                    self.subject.remove_class(class_name, for_binary=self.binary_name)
                elif self.selection_level == ClassSelectionLevel.SubjectLevel:
                    self.subject.remove_class(class_name)
            else:
                if self.selection_level == ClassSelectionLevel.BinaryLevel:
                    self.subject.set_class(class_name, class_value, for_dataset=d, for_binary=self.binary_name)
                elif self.selection_level == ClassSelectionLevel.SubjectLevel:
                    self.subject.set_class(class_name, class_value, for_dataset=d)

        for class_name in class_names:
            class_info = d.get_class(class_name)
            class_box = TClassBox(class_info, callback=change_handler, class_name=class_name)
            self.class_boxes.append(class_box)
            self.layout.addWidget(class_box)

    def set_subject(self, subject: Subject):
        self.subject = subject
        if self.selection_level == ClassSelectionLevel.SubjectLevel:
            self.update_label()

    def set_binary_name(self, binary_name: str):
        self.binary_name = binary_name
        if self.selection_level == ClassSelectionLevel.BinaryLevel:
            self.update_label()

    def set_frame_number(self, frame_number: int):
        self.frame_number = frame_number
        if self.selection_level == ClassSelectionLevel.FrameLevel:
            self.update_label()

    def update_label(self):
        # Update the label to give feedback to the user
        res = f'Selected subject:\n{self.subject.name}\n'
        if self.selection_level == ClassSelectionLevel.BinaryLevel or self.selection_level == ClassSelectionLevel.FrameLevel:
            res += f'Selected binary:\n{self.binary_name}'
        if self.selection_level == ClassSelectionLevel.FrameLevel:
            res += f'Selected frame:\n{self.frame_number}'
        self.label.setText(res)
        # Update the values in the class boxes
        for box in self.class_boxes:
            if self.selection_level == ClassSelectionLevel.BinaryLevel:
                box.update_values(self.subject, for_binary=self.binary_name)
            elif self.selection_level == ClassSelectionLevel.SubjectLevel:
                box.update_values(self.subject)



from tqdm import tqdm
import matplotlib.pyplot as plt
import tensorflow_datasets as tfds
import numpy as np
import tensorflow as tf
from trainer.ml.utils import resize, batcherize, normalize_im
from trainer.ml.models import pix2pix

from trainer.ml import Dataset


# Benchmark the generator
def benchmark(g):
    ts = []
    for t in tqdm(g):
        ts.clear()
        ts.append(t)


# Visualize
def viz():
    vid, gt, frame = next(g)
    print(vid.shape)
    print(gt.shape)
    print(frame)
    im = vid[frame, :, :, :]

    plt.subplot(2, 1, 1)
    plt.imshow(im)
    plt.subplot(2, 1, 2)
    plt.imshow(gt)
    plt.show()


base_model = tf.keras.applications.MobileNetV2(input_shape=[128, 128, 3], include_top=False)

# Use the activations of these layers
layer_names = [
    'block_1_expand_relu',  # 64x64
    'block_3_expand_relu',  # 32x32
    'block_6_expand_relu',  # 16x16
    'block_13_expand_relu',  # 8x8
    'block_16_project',  # 4x4
]
layers = [base_model.get_layer(name).output for name in layer_names]

# Create the feature extraction model
down_stack = tf.keras.Model(inputs=base_model.input, outputs=layers)

down_stack.trainable = False

up_stack = [
    pix2pix.upsample(512, 3),  # 4x4 -> 8x8
    pix2pix.upsample(256, 3),  # 8x8 -> 16x16
    pix2pix.upsample(128, 3),  # 16x16 -> 32x32
    pix2pix.upsample(64, 3),  # 32x32 -> 64x64
]

OUTPUT_CHANNELS = 2
# This is the last layer of the model
last = tf.keras.layers.Conv2DTranspose(
    OUTPUT_CHANNELS, 3, strides=2,
    padding='same', activation='softmax')  # 64x64 -> 128x128

inputs = tf.keras.layers.Input(shape=[128, 128, 3])
x = inputs

# Downsampling through the model
skips = down_stack(x)
x = skips[-1]
skips = reversed(skips[:-1])

# Upsampling and establishing the skip connections
for up, skip in zip(up_stack, skips):
    x = up(x)
    concat = tf.keras.layers.Concatenate()
    x = concat([x, skip])

x = last(x)

model = tf.keras.Model(inputs=inputs, outputs=x)

tfds.disable_progress_bar()
dataset, info = tfds.load('oxford_iiit_pet:3.0.0', with_info=True)


def normalize(input_image, input_mask):
    input_image = tf.cast(input_image, tf.float32) / 255.0
    input_mask -= 1
    return input_image, input_mask


def g_convert(g):
    for vid, gt, f in g:
        if f >= 2:
            res = np.zeros((vid.shape[1], vid.shape[2], 3))
            res[:, :, 0] = vid[f - 2, :, :, 0]
            res[:, :, 1] = vid[f - 1, :, :, 0]
            res[:, :, 2] = vid[f, :, :, 0]
            res_gt = np.zeros((gt.shape[0], gt.shape[1], 1), dtype=np.float32)
            res_gt[:, :, 0] = gt.astype(np.float32)
            yield normalize_im(res).astype(np.float32), res_gt


def g_appendaxis(g):
    for vid, gt in g:
        yield vid, gt[..., np.newaxis]


def get_generator_shape(g_train):
    a, b = next(g_train)
    print(a.shape)
    print(b.shape)
    print(a.dtype)
    print(b.dtype)
    return a, b


def compile_and_train(ds: Dataset, structure_name='bone'):
    g = ds.random_struct_generator(struct_name=structure_name)
    g_extracted = g_convert(g)
    g_resized = resize(g_extracted, (128, 128))
    g_train = batcherize(g_resized, batchsize=1)
    model.compile(optimizer='adam', loss='sparse_categorical_crossentropy',
                  metrics=['accuracy'])
    EPOCHS = 5
    STEPS_PER_EPOCH = 64
    model_history = model.fit_generator(
        g_train,
        epochs=EPOCHS,
        steps_per_epoch=STEPS_PER_EPOCH,
        # verbose=2
    )
    return model_history
# dataset_path, _ = standalone_foldergrab(folder_not_file=True, title='Pick the dataset')
#     ds = Dataset.from_disk(dataset_path)



import itertools
import os
import random
from typing import Dict, Callable

from trainer.ml import Dataset, Subject


def iterate_over_samples(ds: Dataset, f: Callable[[Subject], Subject]):
    """
    Applies a function on every subject in this dataset.
    :param ds: The dataset to be modified
    :param f: f takes a subject and can modify it. The result is automatically saved
    :return:
    """
    for te_name in ds.get_subject_name_list():
        s = ds.get_subject_by_name(te_name)
        s = f(s)
        s.to_disk()


def get_subject_gen(ds: Dataset, split: str = None):
    """
    Iterates once through the dataset. Intended for custom exporting, not machine learning.
    """
    if split is None:
        subjects = ds._json_model["subjects"]
    else:
        subjects = ds._json_model["splits"][split]

    for s_name in subjects:
        yield ds.get_subject_by_name(s_name)


def random_subject_generator(ds: Dataset, split=None):
    if split is None:
        te_names = ds._json_model["subjects"]
    else:
        te_names = ds._json_model["splits"][split]
    random.shuffle(te_names)

    for te_name in itertools.cycle(te_names):
        te = ds.get_subject_by_name(te_name)
        yield te


def random_struct_generator(ds: Dataset, struct_name: str, split=None):
    if split is None:
        subjects = ds._json_model["subjects"]
    else:
        subjects = ds._json_model["splits"][split]

    # Compute the annotated examples for each subject
    annotations: Dict[str, Dict] = {}
    for s_name in subjects:
        s = ds.get_subject_by_name(s_name)
        s_annos = s.get_manual_struct_segmentations(struct_name)
        if s_annos:
            annotations[s_name] = s_annos
    print(annotations)

    for s_name in itertools.cycle(annotations.keys()):
        s = ds.get_subject_by_name(s_name)
        # Randomly pick the frame that will be trained with
        a = annotations[s_name]
        b_name = random.choice(list(annotations[s_name].keys()))
        m_name = random.choice(annotations[s_name][b_name])
        # Build the training example with context
        struct_index = list(s.get_binary_model(b_name)["meta_data"]["structures"].keys()).index(struct_name)
        yield s.get_binary(b_name), s.get_binary(m_name)[:, :, struct_index], \
              s.get_binary_model(m_name)["meta_data"]['frame_number']



"""
This module contains the tooling for:
- CLI tools for training and long file/import/export operations.
"""

import click
import os
from typing import List, Tuple, Callable, Dict

from trainer.ml.data_model import Dataset
from trainer.bib import standalone_foldergrab
from trainer.tools.AnnotationGui import AnnotationGui, run_window
from trainer.bib.JsonClass import dir_is_json_class


@click.group()
def trainer():
    """
    AI command line tools.
    """
    pass


@trainer.group()
def ds():
    """"
    Command line tools concerned with one dataset that is currently on pwd
    """
    pass


@trainer.command(name='init-ds')
@click.option('--parent-path', '-p', default=os.getcwd, help='Directory that the dataset will appear in')
@click.option('--name', '-n', prompt=True, help='Name of the dataset created')
def dataset_init(parent_path, name):
    """
    Create a new dataset
    """
    ls = os.listdir(parent_path)
    click.echo(f"Other datasets in {parent_path}")
    for p in ls:
        if os.path.isdir(p):
            click.echo(f"Dirname: {os.path.basename(p)}")
    if click.confirm(f"The dataset {name} will be created in {parent_path}"):
        d = Dataset.build_new(name, parent_path)
        d.to_disk(parent_path)
    click.echo(f"For working with the dataset {name}, please switch into the directory")


@ds.command(name="annotate")
@click.option('--dataset-path', '-p', default=os.getcwd, help='Path to a dataset')
@click.option('--subject-name', '-s', default='', help='If provided, opens the given subject from the dataset')
def dataset_annotate(dataset_path: str, subject_name: str):
    """
    Start annotating subjects in the dataset.
    """
    if not subject_name:
        # Subject name needs to be picked
        d = Dataset.from_disk(dataset_path)
        subject_name = d.get_subject_name_list()[0]  # Just pick the first subject
    run_window(AnnotationGui, os.path.join(dataset_path, subject_name), dataset_path)


@ds.command(name="server")
@click.option('--dataset-path', '-p', default=os.getcwd, help='Path to a dataset')
def dataset_serve(dataset_path: str):
    print(dataset_path)
    from trainer.server import app
    app.run()


@ds.command(name="train")
@click.option('--dataset-path', '-p', default=os.getcwd, help='The path to the directory where the dataset lives')
def dataset_train(dataset_path: str):
    """
    Trains the models that can be trained from this dataset.

    While True:
        1. Analyses the current subjects for predictable content (structures, classes...)
        2. Determines which models can be trained for helping the annotation process
        3. Cycles over the different models:
            3.1 -> trains them for one epoch
            3.2 -> saves the weights and metadata into the dataset.
    """
    if not dir_is_json_class(dataset_path):
        raise Exception("The given directory is not a valid Dataset")
    d = Dataset.from_disk(dataset_path)

    seg_structs = d.compute_segmentation_structures()

    for str_name in seg_structs:
        click.echo(f"\n{str_name}: {len(seg_structs[str_name])}")
        from trainer.ml.predictor import compile_and_train
        compile_and_train(d, str_name)


@ds.command(name='visualize')
@click.option('--dataset-path', '-p', default=os.getcwd)
@click.option('--subject-name', '-s', default='')
def dataset_visualize(dataset_path: str, subject_name: str):
    ds = Dataset.from_disk(dataset_path)
    if subject_name:
        s = ds.get_subject_by_name(subject_name)
        s.matplot_imagestacks()
    else:
        # TODO Create a html folder with a simple static webview
        # TODO Include Media and Metadata
        raise NotImplementedError()


@ds.command(name='add-image-folder')
@click.option('--dataset-path', '-p', default=os.getcwd)
@click.option('--folder-path', '-ip', default='')
@click.option('--structure-tpl', '-st', default='')
def dataset_add_image_folder(dataset_path: str, folder_path: str, structure_tpl: str):
    ds = Dataset.from_disk(dataset_path)
    if not folder_path:
        folder_path, inputs_dict = standalone_foldergrab(
            folder_not_file=True,
            title='Pick Image folder',
            optional_choices=[('Structure Template', 'str_tpl', ds.get_structure_templates_names())]
        )
        structure_tpl = inputs_dict['str_tpl']
    seg_structs = ds.get_structure_template_by_name(structure_tpl)
    ds.add_image_folder(folder_path, structures=seg_structs)


if __name__ == '__main__':
    trainer()



import os
import shutil
import json
from enum import Enum
from typing import Dict, Callable

import numpy as np

from trainer.bib.enums import BinaryType

BINARIES_DIRNAME = "binaries"
JSON_MODEL_FILENAME = "json_model.json"


def build_full_filepath(name: str, dir_path: str):
    basename = os.path.splitext(name)[0]  # If a file ending is provided it is ignored
    return os.path.join(dir_path, f"{basename}.json")


def dir_is_json_class(dir_name: str, json_checker: Callable[[str], bool] = None):
    if json_checker is None:
        def json_checker(json_val: str) -> bool:
            return True
    # Check if the json exists:
    json_path = os.path.join(dir_name, JSON_MODEL_FILENAME)
    if os.path.exists(json_path):
        with open(json_path, 'r') as f:
            json_content = json.load(f)
        if json_checker(json_content):
            return True
    # TODO Check other constraints (binaries)
    return False


class JsonClass:
    """
    Intended to be subclassed by classes which should to_disk their state
    in the form of numpy binaries (Video, images...) and json metadata (name, example quality...).
    """

    def __init__(self, name: str, model: Dict = None, b_model: Dict = None):
        self.name = name
        self._binaries_dir_path = None

        if model is not None:
            self._json_model = model
        else:
            self._json_model = {}

        if b_model is not None:
            self._binaries_model = b_model
        else:
            self._binaries_model = {}
        self._binaries: Dict[str, np.ndarray] = {}

        self._last_used_parent_dir = None

    @classmethod
    def from_disk(cls, dir_path: str):
        if not dir_is_json_class(dir_path):
            raise Exception(f"{dir_path} is not a valid directory.")

        name = os.path.basename(dir_path)

        full_file_path = os.path.join(dir_path, JSON_MODEL_FILENAME)
        with open(full_file_path, 'r') as f:
            json_content = json.load(f)
        res = cls(name, json_content["payload"], json_content["binaries"])
        res._last_used_parent_dir = os.path.dirname(dir_path)

        # Load binaries
        binaries_dir_path = os.path.join(dir_path, BINARIES_DIRNAME)
        binaries_paths_ls = os.listdir(binaries_dir_path)
        for binary_path in binaries_paths_ls:
            binary_name = os.path.splitext(os.path.basename(binary_path))[0]
            binary_payload = np.load(os.path.join(binaries_dir_path, binary_path), allow_pickle=True)
            res._binaries[binary_name] = binary_payload

        return res

    def to_disk(self, dir_path: str = "", properly_formatted=True, prompt_user=False) -> None:
        if not dir_path:
            dir_path = self._last_used_parent_dir

        # Create the parent directory for this instance
        dir_path = os.path.join(dir_path, self.name)
        if not os.path.exists(dir_path):
            os.mkdir(dir_path)
        file_name = os.path.join(dir_path, JSON_MODEL_FILENAME)
        self._last_used_parent_dir = os.path.dirname(dir_path)

        # Write the json model file
        with open(file_name, 'w+') as f:
            save_json = {
                "payload": self._json_model,
                "binaries": self._binaries_model
            }
            if properly_formatted:
                json.dump(save_json, f, indent=4)
            else:
                json.dump(save_json, f)

        # Write all binaries
        self._binaries_dir_path = os.path.join(dir_path, BINARIES_DIRNAME)
        if not os.path.exists(self._binaries_dir_path):
            os.mkdir(self._binaries_dir_path)
        for binary_key in self._binaries:
            self.save_binary(binary_key)

        if prompt_user:
            os.startfile(file_name)

    def save_binary(self, binary_key):
        if self._binaries_dir_path is None:
            self._binaries_dir_path = os.path.join(self._last_used_parent_dir, BINARIES_DIRNAME)
        np.save(os.path.join(self._binaries_dir_path, binary_key), self._binaries[binary_key])

    def delete_on_disk(self, blocking=True):
        shutil.rmtree(self.get_working_directory(), ignore_errors=True)
        if blocking:
            while os.path.exists(self.get_working_directory()):
                pass

    def get_parent_directory(self):
        return self._last_used_parent_dir

    def get_working_directory(self):
        return os.path.join(self._last_used_parent_dir, self.name)

    def get_json_path(self):
        return os.path.join(self.get_working_directory(), JSON_MODEL_FILENAME)

    def add_binary(self,
                   binary_name: str,
                   binary: np.ndarray,
                   b_type: str = BinaryType.Unknown.value,
                   meta_data=None,
                   auto_save=True):
        # if binary_name in self._binaries:
        #     raise Exception("This binary already exists")
        self._binaries[binary_name] = binary

        self._binaries_model[binary_name] = {
            "numpy_type": str(binary.dtype),
            "binary_type": b_type
        }
        if meta_data is None:
            self._binaries_model[binary_name]["meta_data"] = {}
        else:
            self._binaries_model[binary_name]["meta_data"] = meta_data

        # if auto_save and self._last_used_parent_dir is not None:
        #     self.to_disk(self._last_used_parent_dir)
        self.save_binary(binary_name)

    def remove_binary(self, binary_name):
        # Remove the numpy array from class and disk
        p = os.path.join(self.get_working_directory(), BINARIES_DIRNAME, f"{binary_name}.npy")
        os.remove(p)

        # Remove the key in model
        self._binaries_model.pop(binary_name)
        self._binaries.pop(binary_name)
        self.to_disk(self._last_used_parent_dir)

    def get_binary(self, binary_name):
        return self._binaries[binary_name]

    def get_binary_model(self, binary_name):
        return self._binaries_model[binary_name]

    def get_binary_list_filtered(self, key_filterer: Callable[[Dict], bool]):
        return [i for i in self._binaries_model if key_filterer(self._binaries_model[i])]

    def get_image_stack_keys(self):
        return self.get_binary_list_filtered(lambda x: x["binary_type"] == BinaryType.ImageStack.value)

    def count_binaries(self):
        return len(self._binaries)

    def matplot_imagestacks(self):
        import matplotlib.pyplot as plt
        import seaborn as sns
        for binary in self._binaries_model:
            b = self._binaries[binary]
            btype = self._binaries_model[binary]["binary_type"]
            if btype == BinaryType.ImageStack.value:
                # This is an image or a video
                b = b[0, :, :, :]
                if b.dtype == np.bool:
                    b = b.astype(np.float32) * 255.
                plt.title(binary)
                if b.shape[2] == 1:
                    # Grayscale is assumed
                    sns.heatmap(b[:, :, 0])
                else:
                    plt.imshow(b)
                plt.show()

    def __str__(self):
        res = f"Representation of {self.name}:\n"
        if self._last_used_parent_dir is not None:
            res += f"Last saved at {self.get_working_directory()}\n"
        res += f"Binaries: {len(self._binaries.keys())}\n"
        for binary in self._binaries.keys():
            res += f"{binary}: shape: {self._binaries[binary].shape} (type: {self._binaries[binary].dtype})\n"
            res += f"{json.dumps(self._binaries_model[binary], indent=4)}\n"
        res += json.dumps(self._json_model, indent=4)
        return res

    def __repr__(self):
        return str(self)


if __name__ == '__main__':
    parent_folder = 'C:\\Users\\rapha\\Desktop'
    read = False
    if read:
        example_class = JsonClass.from_disk(os.path.join(parent_folder, "jsonclass_example"))
    else:
        example_class = JsonClass("jsonclass_example", {})
        example_class._json_model["test_number"] = 4
        example_class._json_model["test_list"] = [1, 2, 3]
        example_class._json_model["test_dict"] = {"1": "3", "4": "8"}
        example_class.add_binary("test_image", np.ones((100, 100)))
        example_class.to_disk(parent_folder, prompt_user=True)



from enum import Enum


class MaskType(Enum):
    Unknown = 'unknown'
    Blob = 'blob'
    Point = 'point'
    Line = 'line'


class BinaryType(Enum):
    Unknown = 'unknown'
    ImageStack = 'image_stack'  # Includes videos and volumetric data, shape: (frames, width, height, channels)
    ImageMask = 'img_mask'  # One mask per frame, shape: (width, height, structures)


class ClassType(Enum):
    Binary = 'binary'
    Nominal = 'nominal'
    Ordinal = 'ordinal'


class ClassSelectionLevel(Enum):
    SubjectLevel = "Subject Level"
    BinaryLevel = "Binary Level"
    FrameLevel = "Frame Level"



"""
Produces (smallish) datasets for testing the functionality of the annotator and machine learning functionality.
"""

import os
import random

import numpy as np
import tensorflow as tf
import PySimpleGUI as sg

from trainer.ml import Subject, Dataset
from trainer.bib import load_grayscale_from_disk, standalone_foldergrab, create_identifier, MaskType, ClassType

US_BONE_DATASET = ("https://rwth-aachen.sciebo.de/s/1qO95mdEjhoUBMf/download", "crucial_ligament_diagnosis")


def build_example_te(name: str) -> Subject:
    test_im_path = "../../sample_data/example_image.png"
    test_gt_path = "../../sample_data/example_gt.png"
    te = Subject.build_empty(name)

    im = load_grayscale_from_disk(test_im_path)
    te.add_source_image_by_arr(im, binary_name="src", structures={"bone": MaskType.Line})

    gt = load_grayscale_from_disk(test_gt_path).astype(np.bool)
    te.add_new_gt_by_arr(gt, mask_of="src", structure_names=["bone"])

    from skimage.data import astronaut
    te.add_source_image_by_arr(src_im=astronaut(), binary_name="astronaut_image")

    return te


def create_mnist(p: str, n: str = 'mnist', n_subjects=10):
    """
    Mimics a patient database by solving a classification issue with 10 classes and multiple images per patient.

    Each patient has one of the classes
    - ZERO
    - ONE
    - ...
    - NINE
    :return:
    """
    d = Dataset.build_new(n, p)

    fashion_mnist = tf.keras.datasets.fashion_mnist

    (train_images, train_labels), (test_images, test_labels) = fashion_mnist.load_data()
    class_names = ['TShirtOrTop', 'Trouser', 'Pullover', 'Dress', 'Coat',
                   'Sandal', 'Shirt', 'Sneaker', 'Bag', 'AnkleBoot']

    class_name = "fashion_type"
    d.add_class(class_name, ClassType.Nominal, values=class_names)

    def get_sample_of_class(class_value: str):
        label_indices = np.argwhere(train_labels == class_names.index(class_value))
        sample_index = label_indices[random.randint(0, label_indices.shape[0]), 0]
        return train_images[sample_index, :, :]

    for p_i in range(n_subjects):
        sbjct = Subject.build_empty(f"patient{p_i}")
        d.save_subject(sbjct)
        class_sample = random.choice(class_names)
        for i_i in range(random.randint(1, 3)):
            one_im = get_sample_of_class(class_sample)
            seg_structs = {'outline': MaskType.Line.value,
                           'sleeve': MaskType.Blob.value}
            b_name = create_identifier(hint=f'image{i_i}')
            sbjct.add_source_image_by_arr(one_im, b_name, structures=seg_structs)
            sbjct.set_class(class_name, class_sample, for_dataset=d, for_binary=b_name)
        sbjct.set_class(class_name, class_sample, for_dataset=d)
        sg.OneLineProgressMeter('Creating patient', p_i + 1, n_subjects, 'key',
                                f'Subject: {sbjct.name} of class {class_sample}')
        sbjct.to_disk()
    d.to_disk()


if __name__ == '__main__':
    parent_path, input_keys = standalone_foldergrab(
        folder_not_file=True,
        title="Select the parent folder",
        optional_inputs=[
            ("Name of the dataset", "d_name"),
            ("Max Number Subjects", "#s")
        ],
        optional_choices=[
            ("Dataset Type", "ds_type", ["Fashion Mnist", "Ultrasound"])
        ]
    )
    dataset_name = input_keys['d_name']
    N = input_keys['#s']
    d_path = os.path.join(parent_path, dataset_name)
    if os.path.exists(d_path):
        Dataset.from_disk(d_path).delete_on_disk()
    if input_keys['ds_type'] == 'Fashion Mnist':
        create_mnist(parent_path, dataset_name, n_subjects=int(N))
    elif input_keys['ds_type'] == 'Ultrasound':
        print('Create Ultrasound')
        raise NotImplementedError()



"""
The data model aims to simplify machine learning on complex data structures.
For example classifying a subject (medical patient) by both its gender and between 1 and 4 ultrasound videos.

A dataset contains
- Subjects (Which are the training examples)
- Model Weights
- Config Json files
"""

from __future__ import annotations

import os
import itertools
from tqdm import tqdm
from typing import Callable, List, Dict, Set

import PySimpleGUI as sg
import numpy as np

from trainer.bib import JsonClass, download_and_extract, standalone_foldergrab, create_identifier
from trainer.bib import MaskType, BinaryType, ClassType


class Subject(JsonClass):
    """
    In a medical context a subject is concerned with the data of one patient.
    For example, a patient has classes (disease_1, ...), imaging (US video, CT volumetric data, x-ray image, ...),
    text (symptom description, history) and structured data (date of birth, nationality...).

    Wherever possible the data is saved in json format, but for example for imaging only the metadata is saved
    as json, the actual image file can be found in the binaries-list.

    In future releases a complete changelog will be saved in a format suitable for process mining.
    """

    @classmethod
    def build_empty(cls, name: str):
        res = cls(name=name, model={
            "classes": {}
        })
        return res

    def set_class(self, class_name: str, value: str, for_dataset: Dataset = None, for_binary=""):
        """
        Set a class to true. Classes are stored by their unique string.

        Absence of a class indicates an unknown.

        Hint: If two states of one class can be true to the same time, do not model them as one class.
        Instead of modelling ligament tear as one class, define a binary class for each different ligament.

        :param class_name: Unique string that is used to identify the class.
        :param value: boolean indicating
        :param for_dataset: If provided, set_class checks for compliance with the dataset.
        :param for_binary: If provided, only set the class of the binary and not the whole subject
        :return:
        """
        if for_dataset is not None:
            class_obj = for_dataset.get_class(class_name)
            # print(f"Setting {class_name} to {value}")
            # print(f"{for_dataset.name} tells us about the class:\n{class_obj}")
            if value not in class_obj['values']:
                raise Exception(f"{class_name} cannot be set to {value} according to {for_dataset.name}")

        # Set value
        if for_binary:
            if "classes" not in self._binaries_model[for_binary]["meta_data"]:
                self._binaries_model[for_binary]["meta_data"]["classes"] = {}
            self._binaries_model[for_binary]["meta_data"]["classes"][class_name] = value
        else:
            self._json_model['classes'][class_name] = value

    def get_class_value(self, class_name: str, for_binary=''):
        if for_binary:
            if "classes" not in self._binaries_model[for_binary]["meta_data"]:
                return "--Removed--"
            if class_name in self._binaries_model[for_binary]["meta_data"]["classes"]:
                return self._binaries_model[for_binary]["meta_data"]["classes"][class_name]
        else:
            if class_name in self._json_model['classes']:
                return self._json_model['classes'][class_name]
        return "--Removed--"

    def remove_class(self, class_name: str, for_binary=''):
        if for_binary:
            self._binaries_model[for_binary]["meta_data"]["classes"].pop(class_name)
        else:
            self._json_model['classes'].pop(class_name)

    def add_source_image_by_arr(self,
                                src_im,
                                binary_name: str = "src",
                                structures: Dict[str, str] = None,
                                extra_info: Dict = None):
        """
        Only adds images, not volumes or videos! Unless it is already in shape (frames, width, height, channels).
        Multi-channel images are assumed to be channels last.
        Grayscale images are assumed to be of shape (width, height)
        :param src_im:
        :param binary_name:
        :param structures:
        :param extra_info: Extra info for a human. Must contain only standard types to be json serializable
        :return:
        """
        # Save corresponding json metadata
        meta = {} if structures is None else {"structures": structures}
        if len(src_im.shape) == 2:
            # Assumption: This is a grayscale image
            res = np.reshape(src_im, (1, src_im.shape[0], src_im.shape[1], 1))
            meta["image_type"] = "grayscale"
        elif len(src_im.shape) == 3:
            # This is the image adder function, so assume this is RGB
            res = np.reshape(src_im, (1, src_im.shape[0], src_im.shape[1], src_im.shape[2]))
            meta["image_type"] = "multichannel"
        elif len(src_im.shape) == 4:
            # It is assumed that the array is already in correct shape
            res = src_im.astype(np.uint8) if src_im.dtype != np.uint8 else src_im
            meta["image_type"] = "video"
        else:
            raise Exception("This image can not be an image, check shape!")

        # Extra info
        if extra_info is not None:
            meta["extra"] = extra_info

        self.add_binary(binary_name, res, b_type=BinaryType.ImageStack.value, meta_data=meta)

    def add_file_as_imagestack(self,
                               file_path: str,
                               binary_name='',
                               structures: Dict[str, str] = None):
        """
        Takes an image path and tries to deduce the type of image from the path ending.
        No path ending is assumed to be a DICOM file (not a DICOM folder)
        """
        file_ending = os.path.splitext(file_path)[1]
        if file_ending in ['']:
            from trainer.ml import append_dicom_to_subject
            append_dicom_to_subject(self.get_working_directory(), file_path, binary_name=binary_name,
                                    seg_structs=structures)
        elif file_ending in ['.mp4']:
            print('Video!')
        else:
            raise Exception('This filetype is not understood')

    def delete_gt(self, mask_of: str = None, frame_number=0):
        print(f"Deleting ground truth of {mask_of} at frame {frame_number}")
        gt_name = f"gt_{mask_of}_{frame_number}"  # naming convention
        self.remove_binary(gt_name)

    def add_new_gt_by_arr(self,
                          gt_arr: np.ndarray,
                          structure_names: List[str] = None,
                          mask_of: str = None,
                          frame_number=0):
        """

        :param gt_arr:
        :param structure_names:
        :param mask_of:
        :param frame_number:
        :return: The identifier of this binary
        """
        mismatch_msg = "#structures must correspond to the #channels or be 1 in the case of a single indicated structure"
        assert len(gt_arr.shape) == 2 or gt_arr.shape[2] == len(structure_names), mismatch_msg
        assert gt_arr.dtype == np.bool, "Convert to bool, because the ground truth is assumed to be binary!"
        assert mask_of is not None, "Currently for_src can not be inferred, set a value!"

        if len(gt_arr.shape) == 2:
            # This is a single indicated structure without a last dimension, add it!
            gt_arr = np.reshape(gt_arr, (gt_arr.shape[0], gt_arr.shape[1], 1))

        meta = {
            "mask_of": mask_of,
            "frame_number": frame_number,
            "structures": structure_names
        }
        gt_name = f"gt_{mask_of}_{frame_number}"  # naming convention
        self.add_binary(gt_name, gt_arr, b_type=BinaryType.ImageMask.value, meta_data=meta)

        # TODO set class for this binary if a pixel is non zero in the corresponding binary
        return gt_name

    def get_structure_list(self, image_stack_key: str = ''):
        """
        Computes the possible structures. If no image_stack_key is provided, all possible structures are returned.
        :param image_stack_key:
        :return:
        """
        if image_stack_key:
            if "structures" in self._binaries_model[image_stack_key]["meta_data"]:
                return self._binaries_model[image_stack_key]["meta_data"]["structures"]
            else:
                return []
        else:
            raise NotImplementedError()

    def stack_gts(self, src_name):
        """

        :param src_name: Source binary name
        :return:
        """
        im_shape = self.get_binary(src_name).shape

        def is_gt_of_im(gt: Dict) -> bool:
            if "mask_of" in gt["meta_data"]:
                return gt["meta_data"]["mask_of"] == src_name

        gt_binaries = self.get_binary_list_filtered(is_gt_of_im)
        # print(len(gt_binaries))
        assert len(gt_binaries) == 1  # For a single image there should only be one ground truth
        # res = np.zeros((im_shape[0], im_shape[1], len(gt_binaries) - 1))
        # for channel_index, b_name in enumerate(gt_binaries):
        #     res[:, :, channel_index] = self.get_binary(b_name)
        return self.get_binary(gt_binaries[0])

    def get_grayimage_training_tuple_raw(self, src_name: str):
        training_input = self.get_binary(src_name)
        training_output = self.stack_gts(src_name)
        return training_input[0, :, :, 0].astype(np.float32), training_output.astype(np.float32)

    def get_manual_struct_segmentations(self, struct_name: str) -> Dict[str, List[int]]:
        res = {}

        def filter_imgstack_structs(x: Dict):
            is_img_stack = x['binary_type'] == BinaryType.ImageStack.value
            contains_struct = struct_name in x['meta_data']['structures']
            return is_img_stack and contains_struct

        # Iterate over image stacks that contain the structure
        for b_name in self.get_binary_list_filtered(filter_imgstack_structs):
            # Find the masks of this binary and list them
            bs = []
            for m_name in self.get_binary_list_filtered(
                    lambda x: x['binary_type'] == BinaryType.ImageMask.value and x['meta_data']['mask_of'] == b_name):
                bs.append(m_name)

            if bs:
                res[b_name] = bs
        return res


class Dataset(JsonClass):

    @classmethod
    def build_new(cls, name: str, dir_path: str, example_class=True):
        if os.path.exists(os.path.join(dir_path, name)):
            raise Exception("The directory for this Dataset already exists, use from_disk to load it.")
        res = cls(name, model={
            "subjects": [],
            "splits": {},
            "classes": {},
            "structure_templates": {
                "basic": {"foreground": MaskType.Blob.value,
                          "outline": MaskType.Line.value}
            }
        })
        if example_class:
            res.add_class("example_class", class_type=ClassType.Nominal,
                          values=["Unknown", "Tiger", "Elephant", "Mouse"])
        res.to_disk(dir_path)
        return res

    @classmethod
    def download(cls, url: str, local_path='.', dataset_name: str = None):
        working_dir_path = download_and_extract(url, parent_dir=local_path, dir_name=dataset_name)
        return Dataset.from_disk(working_dir_path)

    def update_weights(self, struct_name: str, weights: np.ndarray):
        print(f"Updating the weights for {struct_name}")
        self.add_binary(struct_name, weights)

    def add_class(self, class_name: str, class_type: ClassType, values: List[str]):
        """
        Adds a class on a dataset level.
        :param class_name:
        :param class_type:
        :param values:
        :return:
        """
        obj = {
            "class_type": class_type.value,
            "values": values
        }
        self._json_model['classes'][class_name] = obj

    def get_class_names(self):
        return list(self._json_model['classes'].keys())

    def get_class(self, class_name: str) -> Dict:
        if class_name in self._json_model['classes']:
            return self._json_model['classes'][class_name]
        else:
            return None

    def remove_class(self, class_name: str):
        self._json_model['classes'].pop(class_name)

    def save_into(self, dir_path: str, properly_formatted=True, prompt_user=False, vis=True) -> None:
        old_working_dir = self.get_working_directory()
        super().to_disk(dir_path, properly_formatted=properly_formatted, prompt_user=prompt_user)
        for i, te_key in enumerate(self._json_model["subjects"]):
            te_path = os.path.join(old_working_dir, te_key)
            te = Subject.from_disk(te_path)
            te.to_disk(self.get_working_directory())
            if vis:
                sg.OneLineProgressMeter('My Meter', i + 1, len(self._json_model['subjects']), 'key',
                                        f'Subject: {te.name}')

    def get_structure_templates_names(self):
        return list(self._json_model["structure_templates"].keys())

    def get_structure_template_by_name(self, tpl_name):
        return self._json_model["structure_templates"][tpl_name]

    def save_subject(self, te: Subject, split=None, auto_save=True):
        # Add the name of the subject into the model
        if te.name not in self._json_model["subjects"]:
            self._json_model["subjects"].append(te.name)

        # Save it as a child directory to this dataset
        te.to_disk(self.get_working_directory())

        if split is not None:
            self.append_subject_to_split(te, split)

        if auto_save:
            self.to_disk(self._last_used_parent_dir)

    def get_subject_name_list(self) -> List[str]:
        return self._json_model["subjects"]

    def append_subject_to_split(self, s: Subject, split: str):
        # Create the split if it does not exist
        if split not in self._json_model["splits"]:
            self._json_model["splits"][split] = []

        self._json_model["splits"][split].append(s.name)

    def append_dataset(self, d_path: str, source_split: str = None, target_split: str = None) -> List:
        """
        Appends the structures from d to this dataset.
        Then copies the subjects.
        :param d_path: Path of the appended dataset
        :param source_split: The split in d that the subjects come from.
        :param target_split: The split that the subjects are appended to.
        :return: The names of the subjects that are copied from d to self
        """
        raise NotImplementedError()  # TODO Update this method
        # Load the other dataset from disk
        d = Dataset.from_disk(d_path)

        # Append structure templates
        self._json_model["structure_templates"].extend(d._json_model["structure_templates"])
        self._json_model["structure_templates"] = list(set(self._json_model["structure_templates"]))

        # Add the subjects
        copied = []
        if source_split is None:
            for te_name in tqdm(d._json_model["subjects"], desc=f"Adding {d.name}"):
                copied.append(te_name)
                self.save_subject(d.get_subject_by_name(te_name), split=target_split, auto_save=False)
        else:
            for te_name in tqdm(d._json_model["splits"][source_split], desc=f"Adding {d.name}"):
                copied.append(te_name)
                self.save_subject(d.get_subject_by_name(te_name), split=target_split, auto_save=False)
        self.to_disk(self._last_used_parent_dir)
        return copied

    def add_image_folder(self, parent_folder: str, structures: Dict[str, str], split=None, progress=True):
        """
        Iterates through a folder.

        If a file is found, a new subject is created with only that file.
        If a directory is found, a new subject is created with all files that live within that directory.
        If a dicom file is found, the image is appended to the subject with that patient_id

        :param parent_folder: Top level folder path
        :param split: The dataset split this data is appended to.
        :param progress: If true, displays a progress bar
        :return:
        """
        top_level_files = os.listdir(parent_folder)
        for i, file_name in enumerate(top_level_files):
            if progress:
                sg.OneLineProgressMeter(
                    title=f'Adding Image Folder',
                    key='key',
                    current_value=i,
                    max_value=len(top_level_files),
                    grab_anywhere=True,
                )

            if os.path.isdir(os.path.join(parent_folder, file_name)):
                pass
            else:  # Assume this is a file
                file_ext = os.path.splitext(os.path.join(parent_folder, file_name))[1]
                if file_ext in ['', '.dcm']:  # Most likely a dicom file
                    from trainer.bib.dicom_utils import import_dicom
                    img_data, meta = import_dicom(os.path.join(parent_folder, file_name))
                    from trainer.bib import slugify
                    p_id = meta['PatientID']
                    p_id_clean = slugify(p_id)
                    if p_id_clean in self.get_subject_name_list():
                        print("load patient")
                        s = Subject.from_disk(os.path.join(self.get_working_directory(), p_id_clean))
                    else:
                        print("Create new patient")
                        s = Subject.build_empty(p_id_clean)
                    s.add_source_image_by_arr(img_data,
                                              create_identifier(hint='DICOM'),
                                              structures=structures,
                                              extra_info=meta)
                    self.save_subject(s, split=split, auto_save=False)

                else:  # Everything else is assumed to be a traditional image file
                    # Create the new subject
                    s_name = os.path.splitext(file_name)[0]
                    s = Subject.build_empty(s_name)
                    self.save_subject(s, split=split, auto_save=False)

                    s.add_file_as_imagestack(os.path.join(parent_folder, file_name), structures=structures)

        self.to_disk()

    def add_ml_folder(self, folder: str, split=None) -> None:
        """
        Assumes a folder structure of the following form:

        - subject 1
            - im (training images)
            - gt1 (segmentation maps for class gt1)
            - gt2 (segmentation maps for class gt2)
            - ...
        - subject 2
            - ...

        The name of the source image and its corresponding ground truths must be identical
        :param split: The training split (train, test...) that the folder is appended to
        :param folder: The path to the parent folder
        :return:
        """
        raise NotImplementedError()  # Update this method
        structure_names = [item for item in os.listdir(folder) if
                           item not in ['im'] and os.path.isdir(os.path.join(folder, item))]
        source_folder = os.path.join(folder, 'im')
        if not os.path.exists(source_folder):
            raise FileNotFoundError("Directory doesnt contain source images")

        source_paths = os.listdir(source_folder)
        for src_path in source_paths:
            te = Subject.build_with_src_image(os.path.splitext(src_path)[0],
                                              os.path.join(source_folder, src_path))
            for structure_name in structure_names:
                structure_path = os.path.join(folder, structure_name)
                gt_p = os.path.join(structure_path, f"{te.name}{os.path.splitext(src_path)[1]}")
                te.add_new_gt_by_path(structure_name, gt_p)
            self.save_subject(te, split=split, auto_save=False)

        self.to_disk()

    def filter_subjects(self, filterer: Callable[[Subject], bool], viz=False) -> List[str]:
        """
        Returns a list with the names of subjects of interest.
        :param filterer: If the filterer returns true, the subject is added to the list
        :return: The list of subjects of interest
        """
        res: List[str] = []
        for i, te_name in enumerate(self._json_model["subjects"]):
            te = self.get_subject_by_name(te_name)
            if filterer(te):
                res.append(te.name)
            if viz:
                sg.OneLineProgressMeter("Filtering subjects", i + 1,
                                        len(self._json_model['subjects']),
                                        'key',
                                        f'Subject: {te.name}')
        return res

    def delete_subjects(self, del_ls: List[Subject]) -> None:
        """
        Deletes a list of subjects
        :param del_ls: List of instances of subjects
        :return:
        """
        for s in tqdm(del_ls, desc="Deleting subjects"):
            del_name = s.name
            s.delete_on_disk()
            self._json_model["subjects"].remove(del_name)
            for split in self._json_model["splits"]:
                if del_name in self._json_model["splits"][split]:
                    self._json_model["splits"][split].remove(del_name)
        self.to_disk(self._last_used_parent_dir)

    def get_subject_by_name(self, te_name: str):
        if te_name not in self._json_model['subjects']:
            raise Exception('This dataset does not contain a subject with this name')
        res = Subject.from_disk(os.path.join(self.get_working_directory(), te_name))
        return res

    def get_summary(self) -> str:
        split_summary = ""
        for split in self._json_model["splits"]:
            split_summary += f"""{split}: {self.__len__(split=split)}\n"""
        return f"Saved at {self.get_working_directory()}\nN: {len(self)}\n{split_summary}"

    def compute_segmentation_structures(self) -> Dict[str, Set[str]]:
        """
        Returns a dictionary.
        Keys: All different structures.
        Values: The names of the subjects that can be used to train these structures with.
        :return: Dictionary of structures and corresponding subjects
        """
        # Segmentation Helper
        seg_structs: Dict[str, Set[str]] = {}  # structure_name: List_of_Training_Example_names with that structure

        def te_filterer(te: Subject) -> bool:
            """
            Can be used to hijack the functional filtering utility
            and uses a side effect of struct_appender to fill seg_structs.
            """

            def struct_appender(b: Dict) -> bool:
                if b['binary_type'] == BinaryType.ImageStack.value:
                    structures = list(b['meta_data']['structures'].keys())
                    for structure in structures:
                        if structure not in seg_structs:
                            seg_structs[structure] = set()
                        if te.name not in seg_structs[structure]:
                            seg_structs[structure] = seg_structs[structure] | {te.name}
                return True

            stacks = te.get_binary_list_filtered(struct_appender)
            return len(stacks) != 0

        self.filter_subjects(lambda x: te_filterer(x))
        return seg_structs

    def get_subject_count(self, split=None):
        if split is None:
            return len(self._json_model["subjects"])
        else:
            return len(self._json_model["splits"][split])

    def __len__(self):
        return self.get_subject_count()

    def __getitem__(self, item):
        return self.get_subject_by_name(item)

    def __iter__(self):
        from trainer.ml.data_loading import get_subject_gen
        return get_subject_gen(self)


if __name__ == '__main__':
    ds_path = download_and_extract(
        online_url='https://rwth-aachen.sciebo.de/s/1qO95mdEjhoUBMf/download',
        parent_dir='./data',  # Your local data folder
        dir_name='crucial_ligament_diagnosis'  # Name of the dataset
    )
    ds = Dataset.from_disk(ds_path)



from trainer.bib.tgui.utils import *
from trainer.bib.tgui.TClassSelector import TClassSelector
from trainer.bib.tgui.TWindow import TWindow, run_window
from trainer.bib.tgui.TSegTool import Brushes, SegToolController



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="ai-trainer",
    version="0.0.2",
    author="Raphael Schaefer",
    author_email="raphaelschaefer1@outlook.com",
    description="AI Trainer",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Telcrome/ai-trainer",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.7',
    download_url='https://github.com/Telcrome/ai-trainer/archive/0.0.2.tar.gz',
    entry_points='''
        [console_scripts]
        trainer=trainer.tools:trainer
    ''',
    install_requires=[
        'numpy',
        'opencv-python',
        'PyQt5',
        'pydicom',
        'PySimpleGui',
    ],
)


"""
The annotation tools takes a subject and allows manual annotation.
For manual annotation the source binary that should be annotated is required.

If a dataset is provided it can be used to select the next subject for annotation,
enabling a conveyor belt like process.
Everything else required should be stored in the subject.
"""

import sys
import os
from typing import Tuple, List, Callable

import numpy as np
import cv2
import imageio
from PyQt5 import QtWidgets, QtCore
import PySimpleGUI as sg

from trainer.bib.tgui import TClassSelector, TWindow, run_window, Brushes, SegToolController
from trainer.ml import Subject, Dataset
from trainer.bib import standalone_foldergrab


def binary_filter(x, for_name, frame_number):
    a = x['binary_type'] == 'img_mask'
    return a and x['meta_data']['mask_of'] == for_name and x['meta_data']['frame_number'] == frame_number


class FrameController(QtWidgets.QWidget):

    def __init__(self, frame_changer: Callable[[int], None]):
        super(FrameController, self).__init__()
        self._current_frame, self._frame_number = -1, -1
        self.frame_changer = frame_changer

        self._layout = QtWidgets.QVBoxLayout()

        self._label = QtWidgets.QLabel("0/0")

        self._slider = QtWidgets.QSlider(QtCore.Qt.Horizontal)
        self._slider.setMinimum(0)
        self._slider.sliderReleased.connect(self.slider_changed)

        self._layout.addWidget(self._label)
        self._layout.addWidget(self._slider)
        self.setLayout(self._layout)

    def slider_changed(self):
        v = self._slider.value()
        self.set_frame_number(v, self._frame_number)
        self.frame_changer(v)

    def set_frame_number(self, i: int, n: int):
        self._current_frame, self._frame_number = i, n
        self._label.setText(f"{self._current_frame}/{self._frame_number}")
        self._slider.setValue(self._current_frame)
        self._slider.setMaximum(self._frame_number)


class AnnotationGui(TWindow):
    """
    Opens a tools window that can be used to fill one subject with values.
    """

    def __init__(self, te_path: str, d_path: str = ""):
        if d_path:
            self.d = Dataset.from_disk(d_path)
        self.current_subject: Subject = None
        self.seg_structs: List = None
        self.frame_number, self.brush, self._struct_index = 0, Brushes.Standard, -1
        self.img_data, self.mask_data, self.mask_name, self._made_changes = None, None, '', False
        self._selected_source_binary, self._selected_gt_binary, self._struct_name = None, None, None

        self.seg_tool = SegToolController(self.selection_event_handler)

        self.main_view = self.seg_tool.get_graphics_scene()
        super().__init__(self.main_view, title="Annotation Window")

        super().init_ui(actions={
            "Subject": [
                super().create_action_from_tuple("Save",
                                                 "Ctrl+S",
                                                 "Saves the currently opened subject on disk",
                                                 self.save_to_disk),
                super().create_action_from_tuple("Add Image Data from Disk",
                                                 "Ctrl+A",
                                                 "Allows to add standard image and video files",
                                                 self.add_image_data_from_disk),
                super().create_action_from_tuple("Add DICOM image data",
                                                 "Ctrl+Shift+A",
                                                 "Note that this assumes a DICOM file with image data",
                                                 self.add_dicom_image_data),
                super().create_action_from_tuple("Delete Selected Binaries",
                                                 "Ctrl+D",
                                                 'Deletes the selected binaries from the source list view',
                                                 self.delete_selected_binaries),
                super().create_action_from_tuple("Delete currently displayed mask",
                                                 "Ctrl+Shift+D",
                                                 'Deletes the selected binaries from the source list view',
                                                 self.delete_selected_mask)
            ],
            "Navigation": [
                super().create_action_from_tuple("Annotate next example",
                                                 "Ctrl+Right",
                                                 "Selects the next example by rules if possible, if not randomly",
                                                 self.select_next_subject),
                super().create_action_from_tuple("Pick next example for annotation",
                                                 "Ctrl+Shift+Right",
                                                 "Opens a subject picker that allows to pick the subject by name",
                                                 self.pick_next_subject),
                super().create_action_from_tuple("Increment displayed frame",
                                                 "Right",
                                                 "Increment displayed frame",
                                                 lambda: self.change_frame(self.frame_number + 1)),
                super().create_action_from_tuple("Decrement displayed frame",
                                                 "Left",
                                                 "Decrement displayed frame",
                                                 lambda: self.change_frame(self.frame_number - 1))
            ],
            "Debug": [
                super().create_action_from_tuple("Inspect dir on Disk",
                                                 "Ctrl+I",
                                                 "Open the directory of the subject on disk",
                                                 self.inspect_on_disk),
                super().create_action_from_tuple("Inspect json on Disk",
                                                 "Ctrl+Shift+I",
                                                 "Open the json file of the subject on disk",
                                                 lambda: self.inspect_on_disk(open_file=True)),
                super().create_action_from_tuple("Update Log",
                                                 "Ctrl+L",
                                                 "Shows the actually loaded json content in the logging tool",
                                                 self.update_log),
                super().create_action_from_tuple("Find frames with annotation",
                                                 "Ctrl+F",
                                                 "Computes the frames with annotations",
                                                 self.find_annotations),
                super().create_action_from_tuple("Ask Model",
                                                 "Ctrl+M",
                                                 "Ask the model",
                                                 self.ask_model)
            ]
        })
        # Frame Controller
        self.frame_controller = FrameController(self.change_frame)
        self.content_grid.add_tool(self.frame_controller)

        # Class selector
        self.class_selector = TClassSelector()
        self.class_selector.configure_selection(self.d)
        self.content_grid.add_tool(self.class_selector)

        # Source binary selector
        self.content_grid.add_tool(QtWidgets.QLabel("Source Binary Selection:"))
        self.lst_source_binaries = QtWidgets.QListWidget()
        self.lst_source_binaries.currentItemChanged.connect(self.lst_src_binaries_changed)
        self.content_grid.add_tool(self.lst_source_binaries)

        # Mask Selector
        self.content_grid.add_tool(QtWidgets.QLabel("Mask Selection:"))
        self.lst_gt_binaries = QtWidgets.QListWidget()
        self.lst_gt_binaries.currentItemChanged.connect(self.lst_gt_binaries_changed)
        self.content_grid.add_tool(self.lst_gt_binaries)

        self.set_current_subject(Subject.from_disk(te_path))

    def ask_model(self):
        from trainer.ml.predictor import predict
        print("The model was consulted")
        channels = self.img_data.shape[-1]
        if self.img_data.shape[-1] == 3:
            gray_image = cv2.cvtColor(self.img_data[self.frame_number, :, :, :], cv2.COLOR_BGR2GRAY)
        else:
            gray_image = self.img_data[self.frame_number, :, :, :]
        struct_name = self.seg_structs[self._struct_index]
        res = predict(gray_image, self.d, struct_name)

        # If no mask exists yet, create a new
        self.create_new_mask()

        self.mask_data[:, :, self._struct_index] = res > 0.3
        self.update()

    def select_next_subject(self):
        self.save_to_disk()

        next_subject_name = self.d.select_for_struct_annotation()
        self.set_current_subject(Subject.from_disk(os.path.join(self.d.get_working_directory(), next_subject_name)))
        print("Selecting next subject")

    def pick_next_subject(self):
        self.save_to_disk()

        tool_width = 20
        tools = [
            [sg.Button(button_text="Annotate subject", key='open_subject', size=(tool_width, 1))]
        ]
        main_layout = [
            [sg.Text(text="Pick subject for annotation", size=(50, 1), key='lbl')],
            [sg.Listbox(key='ls_s',
                        values=[te_name for te_name in self.d.get_subject_name_list()],
                        size=(60, 20)),
             sg.Column(tools)]
        ]

        window = sg.Window("Pick your next subject", layout=main_layout)

        event, values = window.read()
        window.close()

        if event == "open_subject" and values['ls_s']:
            next_subject_name = values['ls_s'][0]
            self.set_current_subject(Subject.from_disk(os.path.join(self.d.get_working_directory(), next_subject_name)))
            print(f"Selecting next subject: {next_subject_name}")

    def save_to_disk(self):
        print(f"Saving to {self.current_subject.get_parent_directory()}")
        self.current_subject.to_disk(self.current_subject.get_parent_directory())

    def add_image_data_from_disk(self):
        layout = [[sg.Text(text="Select from Disk")],
                  [sg.Text(text="Name: "), sg.Input(key='binary_name')],
                  [sg.Input(key='filepath'), sg.FileBrowse()],
                  [sg.Checkbox(text="Reduce to recommended size", key="reduce_size", default=True)],
                  [sg.Submit(), sg.Cancel()]]
        p = sg.Window('Select folder', layout)
        e, v = p.Read()
        p.close()
        if e == "Submit":
            im = imageio.imread(v['filepath'])

            # The annotator cannot deal with big images very well, so resize it if its too big
            if v['reduce_size']:
                dst_pixel = 600
                src_pixel = im.shape[0]
                if dst_pixel < src_pixel:  # Do not resize if the image is small anyways
                    resize_factor = dst_pixel / src_pixel
                    im = cv2.resize(im, None, fx=resize_factor, fy=resize_factor)

            self.current_subject.add_source_image_by_arr(im, binary_name=v['binary_name'], structures={"gt": "blob"})
            self.set_current_subject(self.current_subject)

    def add_dicom_image_data(self):
        dicom_path, ks = standalone_foldergrab(folder_not_file=False,
                                               optional_inputs=[("Binary Name", "binary_name")],
                                               optional_choices=[("Structure Template", "struct_tpl",
                                                                  self.d.get_structure_templates_names())],
                                               title="Select DICOM file")
        if dicom_path:
            from trainer.ml import append_dicom_to_subject
            tpl_name = ks['struct_tpl']
            if tpl_name in self.d.get_structure_templates_names():
                seg_structs = self.d.get_structure_template_by_name(tpl_name)
                append_dicom_to_subject(self.current_subject.get_working_directory(), dicom_path, seg_structs=seg_structs)
                print(ks['binary_name'])
                print(dicom_path)
                self.set_current_subject(Subject.from_disk(self.current_subject.get_working_directory()))

    def inspect_on_disk(self, open_file=False):
        if open_file:
            os.startfile(self.current_subject.get_json_path())
        else:
            os.startfile(self.current_subject.get_working_directory())

    def lst_src_binaries_changed(self, item, auto_save=True):
        if item is not None:
            src_item = item.text()
            print(f"Selected Source binary: {src_item}")
            self.select_source_binary(src_item)

    def lst_gt_binaries_changed(self, item, auto_save=True):
        if item is not None:
            print(f"Selected Ground Truth: {item.text()}")
            self.select_gt_binary(self.seg_structs[self.lst_gt_binaries.currentIndex().row()],
                                  for_name=self._selected_source_binary,
                                  frame_number=self.frame_number)

    def set_current_subject(self, s: Subject):
        self.current_subject = s
        self.frame_number = 0
        self.mask_data = None

        # Handle GUI elements concerned with classes
        self.class_selector.set_subject(self.current_subject)

        # Load the list of source binaries into GUI
        self.lst_source_binaries.clear()
        self.lst_gt_binaries.clear()
        src_names = self.current_subject.get_binary_list_filtered(lambda x: x['binary_type'] == 'image_stack')
        if src_names:
            for b in src_names:
                self.lst_source_binaries.addItem(str(b))
            self.select_source_binary(src_names[0])

        self.update()

    def select_source_binary(self, name: str, auto_save=True):
        if auto_save and self.mask_data is not None and self._made_changes:
            self.save_to_disk()
        self.mask_data, self.mask_name, self.frame_number = None, name, 0

        self.img_data, self._selected_source_binary = self.current_subject.get_binary(name), name
        self.seg_tool.set_img_stack(self.img_data)
        # Load the possible structures into lst_gt
        self.lst_gt_binaries.clear()
        meta = self.current_subject.get_binary_model(name)
        self.seg_structs = list(meta["meta_data"]["structures"].keys())
        for s in self.seg_structs:
            self.lst_gt_binaries.addItem(f'{s}: {meta["meta_data"]["structures"][s]}')

        # Preselect a mask if there is only one structure anyway
        if self.current_subject.get_binary_list_filtered(lambda x: binary_filter(x, name, self.frame_number)):
            self.select_gt_binary(self.seg_structs[0], for_name=name)
        else:
            self._selected_gt_binary, self.mask_data = None, None

        # Inform the class selector about the new selected binary
        self.class_selector.set_binary_name(name)

        self.update()

    def select_gt_binary(self, structure_name, for_name: str, frame_number=0, auto_create=False, auto_save=True):
        """
        For every source binary the subject knows which structures can exist by template.
        This function looks for the ground truth and if there is none creates a new fitting one.
        :param structure_name: The selected structure name
        :param for_name: The source binary which the structure should correspond to
        :param frame_number: The frame number which is selected for annotation
        :param auto_create: Automatically write a new ground truth array on disk if there is none
        :param auto_save: Automatically save the previously selected gt and binary first
        :return:
        """

        # See if this structure already is described in one of the masks of this frame number
        gt_names = self.current_subject.get_binary_list_filtered(lambda x: binary_filter(x, for_name, frame_number))

        # Currently assume that only one masks exists for every frame, which describes all structs
        assert len(gt_names) <= 1

        def load_mask(gt_name: str) -> Tuple[str, str, int, np.ndarray]:
            """
            Loads the right mask out of the subject.
            :param gt_name: The name of the ground truth binary that contains masks of all structures.
            :return:
            """
            gt = self.current_subject.get_binary_model(gt_name)
            gt_arr = self.current_subject.get_binary(gt_name)
            i = gt['meta_data']['structures'].index(structure_name)
            return gt_name, structure_name, i, gt_arr

        if gt_names:
            self._selected_gt_binary, self._struct_name, self._struct_index, self.mask_data = load_mask(gt_names[0])
        elif auto_create:
            # Create a new ground truth for this frame
            src_b = self.current_subject.get_binary(for_name)
            new_gt_name = self.current_subject.add_new_gt_by_arr(
                np.zeros((src_b.shape[1], src_b.shape[2], len(self.seg_structs)), dtype=np.bool),
                structure_names=self.seg_structs,
                mask_of=for_name,
                frame_number=self.frame_number)
            self._selected_gt_binary, self._struct_name, self._struct_index, self.mask_data = load_mask(new_gt_name)
        else:
            self._selected_gt_binary, self.mask_data = None, None

        self.update()
        self._made_changes = False

    def update_log(self):
        # self.content_grid.logging_widget.clear()
        # self.content_grid.logging_widget.appendPlainText(f"{str(self.te)}")
        sg.popup(str(self.current_subject))

    def delete_selected_binaries(self):
        ns = [i.text() for i in self.lst_source_binaries.selectedItems()]
        for n in ns:
            print(f"Deleting binary {n}")
            self.current_subject.remove_binary(n)
        self.set_current_subject(self.current_subject)

    def delete_selected_mask(self):
        self.current_subject.delete_gt(mask_of=self._selected_source_binary, frame_number=self.frame_number)
        self._selected_gt_binary, self.mask_data = None, None

        self.update()
        self._made_changes = False

    def selection_event_handler(self, ps: List[Tuple[int, int]], add=True):
        if not self.mask_name:
            sg.popup("Please set a mask name")

        for p in ps:
            self.add_point(p, add=add)

        self.seg_tool.set_mask(self.mask_data, self.current_subject.get_binary_model(self._selected_source_binary))
        self.seg_tool.display_mask(self._struct_name)
        self._made_changes = True

    def change_frame(self, frame_number: int):
        if self.img_data.shape[0] - 1 >= frame_number >= 0:
            self.frame_number = frame_number

            self.select_gt_binary(self._struct_name,
                                  for_name=self._selected_source_binary,
                                  frame_number=frame_number)

            self.update()
        else:
            msg = f"Cannot set frame {self.frame_number} on multi frame image with {self.img_data.shape[0]} frames"
            print(msg)

    def create_new_mask(self):
        if self.mask_data is None:
            self.select_gt_binary(
                self.seg_structs[self.lst_gt_binaries.currentIndex().row()],
                for_name=self._selected_source_binary,
                frame_number=self.frame_number,
                auto_create=True)

    def add_point(self, pos, add=True):
        # If the ground truth does not yet exist, create a new
        self.create_new_mask()

        if self.brush == Brushes.Standard:
            intensity = add * 255
            tmp = cv2.circle(self.mask_data[:, :, self._struct_index].astype(np.uint8) * 255, pos,
                             self.seg_tool.pen_size, intensity, -1)
            self.mask_data[:, :, self._struct_index] = tmp.astype(np.bool)
        # elif self.m.brush == Brushes.AI_Merge:
        #     intensity = add * 255
        #     c = cv2.circle(np.zeros(self.m.foreground.shape, dtype=np.uint8), pos, self.m.brush_size, intensity, -1)
        #     tmp = (c & self.m.proposal) | self.m.foreground
        #     self.m.foreground = tmp.astype(np.bool)

    def update(self):
        self.seg_tool.display_img_stack(frame_number=self.frame_number)
        self.seg_tool.set_mask(self.mask_data,
                               self.current_subject.get_binary_model(self._selected_source_binary))
        self.seg_tool.display_mask(self._struct_name)
        self.frame_controller.set_frame_number(self.frame_number, self.img_data.shape[0])

    def find_annotations(self):
        struct_name = self.seg_structs[self._struct_index]
        annos = self.current_subject.get_manual_struct_segmentations(struct_name)
        if self._selected_source_binary in annos:
            bs = annos[self._selected_source_binary]
            res = f"{struct_name} is segmented in {len(bs)} frames:\n"
            f_str = [f"{self.current_subject.get_binary_model(b_name)['meta_data']['frame_number']}\n" for b_name in bs]
            for f in f_str:
                res += f
            sg.Popup(res)
        else:
            sg.Popup(f"No Annotations made for {struct_name}")


if __name__ == '__main__':
    print(sys.argv)
    if len(sys.argv) < 2:
        folder_path = standalone_foldergrab(folder_not_file=True,
                                            title="Please select the subject that you want to add")
        run_window(AnnotationGui, folder_path)
    else:
        run_window(AnnotationGui, sys.argv[1])
